#
#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

import ast
import logging
import random
import time
import warnings

from gevent.event import AsyncResult

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.GenericDiffractometer import GenericDiffractometer


class LNLSDiffractometer(GenericDiffractometer):
    """
    Descript. :
    """

    def __init__(self, *args):
        """
        Descript. :
        """
        GenericDiffractometer.__init__(self, *args)

    def init(self):
        """
        Descript. :
        """
        # self.image_width = 100
        # self.image_height = 100

        GenericDiffractometer.init(self)
        # Bzoom: 1.86 um/pixel (or 0.00186 mm/pixel) at minimum zoom
        self.x_calib = 0.00186
        self.y_calib = 0.00186
        self.last_centred_position = [318, 238]

        self.pixels_per_mm_x = 1.0 / self.x_calib
        self.pixels_per_mm_y = 1.0 / self.y_calib
        if "zoom" not in self.motor_hwobj_dict.keys():
            self.motor_hwobj_dict["zoom"] = self.get_object_by_role("zoom")
        calibration_x = self.zoom.getProperty("mm_per_pixel_x")
        calibration_y = self.zoom.getProperty("mm_per_pixel_y")
        self.zoom_calibration_x = ast.literal_eval(calibration_x)
        self.zoom_calibration_y = ast.literal_eval(calibration_y)

        self.beam_position = [318, 238]

        self.current_phase = GenericDiffractometer.PHASE_CENTRING

        self.cancel_centring_methods = {}
        self.current_motor_positions = {
            "phiy": 1.0,
            "sampx": 0.0,
            "sampy": -1.0,
            "zoom": 8.53,
            "focus": -0.42,
            "phiz": 1.1,
            "phi": 311.1,
            "kappa": 11,
            "kappa_phi": 22.0,
        }
        #self.move_motors(self._get_random_centring_position())

        self.current_state_dict = {}
        self.centring_status = {"valid": False}
        self.centring_time = 0

        # self.image_width = 400
        # self.image_height = 400

        self.mount_mode = self.getProperty("sample_mount_mode")
        if self.mount_mode is None:
            self.mount_mode = "manual"

        self.equipment_ready()

        # TODO FFS get this cleared up - one function, one name
        self.getPositions = self.get_positions
        #self.moveMotors = self.move_motors

        self.connect(
            self.motor_hwobj_dict["phi"], "positionChanged", self.phi_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiy"], "positionChanged", self.phiy_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiz"], "positionChanged", self.phiz_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa"], "positionChanged", self.kappa_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa_phi"],
            "positionChanged",
            self.kappa_phi_motor_moved,
        )
        self.connect(
            self.motor_hwobj_dict["sampx"], "positionChanged", self.sampx_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["sampy"], "positionChanged", self.sampy_motor_moved
        )

    def getStatus(self):
        """
        Descript. :
        """
        return "ready"

    def execute_server_task(self, method, timeout=30, *args):
        return

    def in_plate_mode(self):
        return self.mount_mode == "plate"

    def use_sample_changer(self):
        return self.mount_mode == "sample_changer"

    def is_reversing_rotation(self):
        return True

    def get_grid_direction(self):
        """
        Descript. :
        """
        return self.grid_direction

    def manual_centring(self):
        """
        Descript. :
        """
        for click in range(3):
            self.user_clicked_event = AsyncResult()
            x, y = self.user_clicked_event.get()
            if click < 2:
                self.motor_hwobj_dict["phi"].set_value_relative(90)
        self.last_centred_position[0] = x
        self.last_centred_position[1] = y
        centred_pos_dir = self._get_random_centring_position()
        return centred_pos_dir

    def automatic_centring(self):
        """Automatic centring procedure"""
        centred_pos_dir = self._get_random_centring_position()
        self.emit("newAutomaticCentringPoint", centred_pos_dir)
        return centred_pos_dir

    def _get_random_centring_position(self):
        """Get random centring result for current positions"""

        # Names of motors to vary during centring
        vary_motor_names = ("sampx", "sampy", "phiy")

        # Range of random variation
        var_range = 0.08

        # absolute value limit for varied motors
        var_limit = 2.0

        result = self.current_motor_positions.copy()
        for tag in vary_motor_names:
            val = result.get(tag)
            if val is not None:
                random_num = random.random()
                var = (random_num - 0.5) * var_range
                val += var
                if abs(val) > var_limit:
                    val *= 1 - var_range / var_limit
                result[tag] = val
        return result

    def is_ready(self):
        """
        Descript. :
        """
        return True

    def isValid(self):
        """
        Descript. :
        """
        return True

    def invalidate_centring(self):
        """
        Descript. :
        """
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            # self.emitProgressMessage("")
            self.emit("centringInvalid", ())

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """
        Descript. :
        """
        centred_pos_dir = self._get_random_centring_position()
        return centred_pos_dir

    def get_calibration_data(self, offset):
        """
        Descript. :
        """
        # return (1.0 / self.x_calib, 1.0 / self.y_calib)
        return (1.0 / self.x_calib, 1.0 / self.y_calib)

    def refresh_omega_reference_position(self):
        """
        Descript. :
        """
        return

    # def get_omega_axis_position(self):
    #     """
    #     Descript. :
    #     """
    #     return self.current_positions_dict.get("phi")

    def beam_position_changed(self, value):
        """
        Descript. :
        """
        self.beam_position = value

    def get_current_centring_method(self):
        """
        Descript. :
        """
        return self.current_centring_method

    def motor_positions_to_screen(self, centred_positions_dict):
        """
        Descript. :
        """
        return self.last_centred_position[0], self.last_centred_position[1]

    def moveToCentredPosition(self, centred_position, wait=False):
        """
        Descript. :
        """
        try:
            return self.move_to_centred_position(centred_position)
        except BaseException:
            logging.exception("Could not move to centred position")

    def phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["phi"] = pos
        self.emit("phiMotorMoved", pos)

    def phiy_motor_moved(self, pos):
        self.current_motor_positions["phiy"] = pos

    def phiz_motor_moved(self, pos):
        self.current_motor_positions["phiz"] = pos

    def sampx_motor_moved(self, pos):
        self.current_motor_positions["sampx"] = pos

    def sampy_motor_moved(self, pos):
        self.current_motor_positions["sampy"] = pos

    def kappa_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaMotorMoved", pos)

    def kappa_phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa_phi"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaPhiMotorMoved", pos)

    def refresh_video(self):
        """
        Descript. :
        """
        self.emit("minidiffStateChanged", "testState")
        if HWR.beamline.beam:
            HWR.beamline.beam.beam_pos_hor_changed(300)
            HWR.beamline.beam.beam_pos_ver_changed(200)

    def start_auto_focus(self):
        """
        Descript. :
        """
        return

    def calculate_move_to_beam_pos(self, x, y):
        """
        Descript. : calculate motor positions to put sample on the beam.
        Returns: dict of motor positions
        """
        # Update beam position
        self.beam_position[0], self.beam_position[1] = HWR.beamline.beam.get_beam_position_on_screen()

        print(("moving to beam position: %d %d" % (
            self.beam_position[0],
            self.beam_position[1],
        )))

        # Set velocity of omega to move during centring
        #self.set_omega_default_velocity()

        # Set scale of pixels per mm according to current zoom
        #self.pixels_per_mm_x = self.motor_zoom_hwobj.getPixelsPerMm(0)
        #self.pixels_per_mm_y = self.motor_zoom_hwobj.getPixelsPerMm(1)

        # Get clicked position of mouse pointer
        #self.user_clicked_event = AsyncResult()
        #x, y = self.user_clicked_event.get()
        # Last clicked position
        self.last_centred_position[0] = x
        self.last_centred_position[1] = y

        # Get current value of involved motors
        omega_pos  = self.motor_hwobj_dict["phi"].get_value()
        # For now, phiz refers to gonio x motor
        goniox_pos = self.motor_hwobj_dict["phiz"].get_value()
        sampx_pos  = self.motor_hwobj_dict["sampx"].get_value()
        sampy_pos  = self.motor_hwobj_dict["sampy"].get_value()

        # Pixels to move axis X of whole goniometer
        import math
        drx_goniox = abs(- x - y + 1152) / math.sqrt(2)
        if y <= (- x + 1152):
            dir_goniox = 1
        else:
            dir_goniox = -1
        move_goniox = dir_goniox * drx_goniox
        # mm to move
        move_goniox = move_goniox / self.pixels_per_mm_x

        # Move absolute
        move_goniox += goniox_pos

        # Calculate new position of X
        dry_samp = abs(x - y - 128) / math.sqrt(2)
        if (y >= x - 128):
            dir_samp = 1
        else:
            dir_samp = -1
        move_samp = dir_samp * dry_samp
        print('move_samp = ' + str(move_samp))

        move_sampx = (math.sin(math.radians(omega_pos)) * move_samp)
        # print("math.cos(math.radians(omega_pos)): ", math.cos(math.radians(omega_pos)))
        # print("self.beam_position[1]: ", self.beam_position[1])
        # print("float(last_centred_position[1])", float(last_centred_position[1]))
        # print("move_sampx = (math.cos(math.radians(omega_pos)) * (self.beam_position[1] - float(last_centred_position[1]))): ", move_sampx)
        #move_sampx = move_sampx / self.pixels_per_mm_x
        move_sampx = (move_sampx / self.pixels_per_mm_x) * -1
        # print("move_sampx = move_sampx / self.pixels_per_mm_x: ", move_sampx)
        # Move absolute
        move_sampx += sampx_pos
        # print("move_sampx += sampx_pos: ", move_sampx)

        # Calculate new position of Y
        move_sampy = (math.cos(math.radians(omega_pos)) * move_samp)
        # print("math.sin(math.radians(omega_pos)): ", math.sin(math.radians(omega_pos)))
        # print("self.beam_position[1]: ", self.beam_position[1])
        # print("float(last_centred_position[1])", float(last_centred_position[1]))
        # print("move_sampy = (math.sin(math.radians(omega_pos)) * (self.beam_position[1] - float(last_centred_position[1]))): ", move_sampy)
        move_sampy = (move_sampy / self.pixels_per_mm_y)
        print('move_sampy = ' + str(move_sampy))
        #move_sampy = move_sampy / self.pixels_per_mm_y
        # print("move_sampy = move_sampy / self.pixels_per_mm_y: ", move_sampy)
        # Move absolute
        move_sampy += sampy_pos
        # print("move_sampy += sampy_pos: ", move_sampy)
        centred_pos_dir = { 'phiz': move_goniox, 'sampx': move_sampx, 'sampy': move_sampy }
        print('Target pos = ' + str(centred_pos_dir))
        return centred_pos_dir

    def move_to_beam(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """

        centred_pos_dir = self.calculate_move_to_beam_pos(x, y)
        print('Moving motors to beam...')
        self.move_to_motors_positions(centred_pos_dir, wait=True)
        return centred_pos_dir

    def move_to_coord(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """
        warnings.warn(
            "Deprecated method, call move_to_beam instead", DeprecationWarning
        )
        return self.move_to_beam(x, y, omega)

    def start_move_to_beam(self, coord_x=None, coord_y=None, omega=None):
        """
        Descript. :
        """
        self.last_centred_position[0] = coord_x
        self.last_centred_position[1] = coord_y
        self.centring_time = time.time()
        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centring_status = {
            "valid": True,
            "startTime": curr_time,
            "endTime": curr_time,
        }
        motors = self.get_positions()
        motors["beam_x"] = 0.1
        motors["beam_y"] = 0.1
        self.last_centred_position[0] = coord_x
        self.last_centred_position[1] = coord_y
        self.centring_status["motors"] = motors
        self.centring_status["valid"] = True
        self.centring_status["angleLimit"] = False
        self.emit_progress_message("")
        self.accept_centring()
        self.current_centring_method = None
        self.current_centring_procedure = None

    def update_values(self):
        self.emit("zoomMotorPredefinedPositionChanged", None, None)
        omega_ref = [0, 238]
        self.emit("omegaReferenceChanged", omega_ref)

    def move_kappa_and_phi(self, kappa, kappa_phi):
        return

    def get_osc_max_speed(self):
        return 66

    def get_osc_limits(self):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_osc_dynamic_limits(self):
        """Returns dynamic limits of oscillation axis"""
        return (0, 20)

    def get_scan_dynamic_limits(self, speed=None):
        return (-360, 360)

    def move_omega_relative(self, relative_angle):
        self.motor_hwobj_dict["phi"].syncMoveRelative(relative_angle, 5)

    def set_phase(self, phase, timeout=None):
        self.current_phase = str(phase)
        self.emit("minidiffPhaseChanged", (self.current_phase,))

    def get_point_from_line(self, point_one, point_two, index, images_num):
        return point_one.as_dict()

    @property
    def zoom(self):
        """
        Override method.
        """
        return self.motor_hwobj_dict.get("zoom")

    def get_zoom_calibration(self):
        """Returns tuple with current zoom calibration (px per mm)."""
        zoom_enum = self.zoom.get_value()  # Get current zoom enum
        zoom_enum_str = zoom_enum.name # as str
        self.x_calib = self.zoom_calibration_x.get(zoom_enum_str)
        self.y_calib = self.zoom_calibration_y.get(zoom_enum_str)
        try:
            float(self.x_calib)
            float(self.y_calib)
            self.pixels_per_mm_x = 1.0 / self.x_calib
            self.pixels_per_mm_y = 1.0 / self.y_calib
        except Exception as e:
            print("[Zoom] Error on calibration: " + str(e))
        return (self.pixels_per_mm_x, self.pixels_per_mm_y)

    def get_pixels_per_mm(self):
        """
        Override method.
        """
        pixels_per_mm_x, pixels_per_mm_y = self.get_zoom_calibration()
        return (pixels_per_mm_x, pixels_per_mm_y)

    def update_zoom_calibration(self):
        """
        Override method.
        """
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x,
                                          self.pixels_per_mm_y)))



from datetime import datetime

from mxcubecore import HardwareRepository as HWR

from .ESRFEnergyScan import ESRFEnergyScan


class ID29EnergyScan(ESRFEnergyScan):

    def energy_scan_hook(self, energy_scan_parameters):
        self.energy = energy_scan_parameters["edgeEnergy"]
        if self.energy_scan_parameters["findattEnergy"]:
            HWR.beamline.energy.set_value(energy_scan_parameters["findattEnergy"])

    def set_mca_roi(self, eroi_min, eroi_max):
        """
        self.mca = self.get_object_by_role("MCA")
        self.energy_scan_parameters["fluorescenceDetector"] = self.mca.get_property(
            "username"
        )
        """
        self.energy_scan_parameters["fluorescenceDetector"] = "KETEK_AXAS-A"
        # check if roi in eV or keV
        if eroi_min > 1000:
            eroi_min /= 1000.0
            eroi_max /= 1000.0
        self.ctrl.mca.set_roi(
            eroi_min,
            eroi_max,
            channel=1,
            element=self.energy_scan_parameters["element"],
            atomic_nb=self.energy_scan_parameters["atomic_nb"],
        )
        print(self.ctrl.mca.get_roi())

    def choose_attenuation(self):
        eroi_min = self.energy_scan_parameters["eroi_min"]
        eroi_max = self.energy_scan_parameters["eroi_max"]
        self.ctrl.detcover.set_in()
        self.ctrl.find_max_attenuation(ctime=2, roi=[eroi_min, eroi_max], datafile="/tmp/abb")
        self.energy_scan_parameters[
            "transmissionFactor"
        ] = HWR.beamline.transmission.get_value()

    def execute_energy_scan(self, energy_scan_parameters):
        startE = energy_scan_parameters["startEnergy"]
        endE = energy_scan_parameters["endEnergy"]
        dd = datetime.now()
        fname = "%s/%s_%s_%s_%s.scan" % (
            energy_scan_parameters["directory"],
            energy_scan_parameters["prefix"],
            datetime.strftime(dd, "%d"),
            datetime.strftime(dd, "%B"),
            datetime.strftime(dd, "%Y"),
        )
        self.ctrl.energy_scan.do_energy_scan(startE, endE, datafile=fname)

        """
        self.energy_scan_parameters["exposureTime"] = self.ctrl.MONOSCAN_INITSTATE[
            "exposure_time"
        ]
        """

    def escan_prepare(self):
        self.ctrl = self.get_object_by_role("controller")

        self.ctrl.detcover.set_in()
        self.ctrl.diffractometer.fldet_in()
        #self.ctrl.fluodet.IN
        self.ctrl.diffractometer.set_phase("DataCollection")

        if self.beamsize:
            # get the aperture size
            bsX = self.beamsize.get_size(self.beamsize.get_value().name)
            self.energy_scan_parameters["beamSizeHorizontal"] = bsX
            self.energy_scan_parameters["beamSizeVertical"] = bsX



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"
import logging
import math
import os
import pickle
import sys
import time
from enum import (
    Enum,
    unique,
)

import gevent
import lmfit
import numpy as np
import sample_centring
import simplejpeg
from gevent.event import AsyncResult
from tango import (
    DevFailed,
    DeviceProxy,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.GenericDiffractometer import (
    DiffractometerState,
    GenericDiffractometer,
)
from mxcubecore.TaskUtils import task

murko_path = os.getenv("MURKO_PATH")
sys.path.insert(1, murko_path)
from murko import (
    get_predictions,
    plot_analysis,
)

if sys.version_info[0] >= 3:
    unicode = str


@unique
class PhaseStates(Enum):
    MOVING = "moving"
    READY = "ready"


class P11NanoDiff(GenericDiffractometer):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        super().__init__(name)

        self.PHASE_STATES = PhaseStates

        self.detcover_hwobj = None
        self.collimator_hwobj = None
        self.beamstop_hwobj = None
        self.backlight_hwobj = None
        self.yag_hwobj = None
        self.pinhole_hwobj = None

        self.ignore_pinhole = True

        self.ai_finished = False

    def init(self):
        """
        Descript. :
        """

        self.diffractometer_state = DiffractometerState.Unknown

        self.current_phase = GenericDiffractometer.PHASE_UNKNOWN
        self.phase_goingto = None
        self.moving_motors = False
        self.phase_state = self.PHASE_STATES.READY

        self.cancel_centring_methods = {}

        self.current_state_dict = {}
        self.centring_status = {"valid": False}
        self.centring_time = 0

        self.save_motor_list = None
        self.pixels_per_mm_x, self.pixels_per_mm_y = (None, None)

        self._saved_position = {}
        self._saved_position["transfer"] = {
            "phix": 0,
            "phiy": 0,
            "phiz": 0,
            "microy": 0,
            "microz": 0,
            "sampx": 0,
            "sampy": 0,
        }

        save_motors = self.get_property("save_motors")

        self.log.debug("SAVE MOTORS are: %s" % str(save_motors))

        if save_motors:
            self.save_motor_list = [
                motname.strip() for motname in save_motors.split(",")
            ]

        GenericDiffractometer.init(self)

        # using sample_centring module
        self.centring_sampx = sample_centring.CentringMotor(
            self.motor_hwobj_dict["sampx"], units="microns"
        )
        self.centring_sampy = sample_centring.CentringMotor(
            self.motor_hwobj_dict["sampy"], units="microns"
        )

        self.centring_phi = sample_centring.CentringMotor(
            self.motor_hwobj_dict["phi"], direction=-1
        )
        self.centring_phiz = sample_centring.CentringMotor(
            self.motor_hwobj_dict["phiz"], direction=1, units="microns"
        )
        self.centring_phiy = sample_centring.CentringMotor(
            self.motor_hwobj_dict["phiy"], direction=1, units="microns"
        )

        self.detcover_hwobj = self.get_object_by_role("detector-cover")
        self.collimator_hwobj = self.get_object_by_role("collimator")
        self.beamstop_hwobj = self.get_object_by_role("beamstop")
        self.yag_hwobj = self.get_object_by_role("yag")
        self.pinhole_hwobj = self.get_object_by_role("pinhole")
        self.backlight_hwobj = self.get_object_by_role("backlight")

        self.omega_hwobj = self.motor_hwobj_dict["phi"]

        self.connect(self.detcover_hwobj, "valueChanged", self.update_phase)
        self.connect(self.backlight_hwobj, "valueChanged", self.update_phase)
        self.connect(self.collimator_hwobj, "valueChanged", self.update_phase)
        self.connect(self.yag_hwobj, "valueChanged", self.update_phase)
        self.connect(self.beamstop_hwobj, "valueChanged", self.update_phase)

        self.connect(self.omega_hwobj, "stateChanged", self.update_phase)

        self.update_phase()
        self.update_zoom_calibration()

        self.update_beam_position()
        self.nclicks = 3
        self.step = 120

        self.lower_bound_ch = self.get_channel_object("acq_lower_bound")
        self.upper_bound_ch = self.get_channel_object("acq_upper_bound")

        self.acq_arm_cmd = self.get_command_object("acq_arm")
        self.acq_abort = self.get_command_object("acq_abort")
        self.acq_on_cmd = self.get_command_object("acq_on")
        self.acq_off_cmd = self.get_command_object("acq_off")
        self.acq_window_off_cmd = self.get_command_object("acq_window_off")

        if None in [
            self.lower_bound_ch,
            self.upper_bound_ch,
            self.acq_arm_cmd,
            self.acq_on_cmd,
            self.acq_off_cmd,
            self.acq_window_off_cmd,
        ]:
            self.init_ok = False
            self.log.debug("lower_bound_ch: %s" % self.lower_bound_ch)
            self.log.debug("upper_bound_ch: %s" % self.upper_bound_ch)
            self.log.debug("acq_arm_cmd: %s" % self.acq_arm_cmd)
            self.log.debug("acq_on_cmd: %s" % self.acq_on_cmd)
            self.log.debug("acq_off_cmd: %s" % self.acq_off_cmd)
            self.log.debug("acq_window_off_cmd: %s" % self.acq_window_off_cmd)
        else:
            self.init_ok = True


    def update_beam_position(self):
        zoom_hwobj = self.motor_hwobj_dict["zoom"]
        image_dimensions = zoom_hwobj.camera_hwobj.get_image_dimensions()
        self.beam_position = [image_dimensions[0] / 2, image_dimensions[1] / 2]
        self.zoom_centre["x"] = self.beam_position[0]
        self.zoom_centre["y"] = self.beam_position[1]

    def update_zoom_calibration(self):
        zoom_hwobj = self.motor_hwobj_dict["zoom"]
        pixels_per_mm_x, pixels_per_mm_y = zoom_hwobj.get_pixels_per_mm()
        if (pixels_per_mm_x != self.pixels_per_mm_x) or (
            pixels_per_mm_y != self.pixels_per_mm_y
        ):
            self.pixels_per_mm_x, self.pixels_per_mm_y = (
                pixels_per_mm_x,
                pixels_per_mm_y,
            )
            self.emit(
                "pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),)
            )

    def execute_server_task(self, method, timeout=30, *args):
        return

    def is_reversing_rotation(self):
        return True

    def get_grid_direction(self):
        """
        Descript. :
        """
        return self.grid_direction

    def predict_click_pix(self, frame):

        request_arguments = {}
        request_arguments["to_predict"] = frame
        image_jpeg = request_arguments["to_predict"]
        image_jpeg = simplejpeg.decode_jpeg(image_jpeg)
        request_arguments["description"] = [
            "foreground",
            "crystal",
            "loop_inside",
            "loop",
            ["crystal", "loop"],
            ["crystal", "loop", "stem"],
        ]
        request_arguments["save"] = False
        request_arguments["prefix"] = "predicted"
        _start = time.time()

        # Select here the host and port where murko server is running
        murko_host = os.getenv("MURKO_HOST")
        murko_port = int(os.getenv("MURKO_PORT"))
        analysis = get_predictions(request_arguments, host=murko_host, port=murko_port)

        original_image_shape = analysis["original_image_shape"]
        sizeOfPictureY, sizeOfPictureX = original_image_shape[:2]
        description = analysis["descriptions"][0]
        self.log.debug(
            "Client got all predictions in %.4f seconds" % (time.time() - _start)
        )
        anything_in_the_picture = description["present"]

        if anything_in_the_picture:
            loop_present, r, c, h, w = description["aoi_bbox"]
            if loop_present:
                self.log.debug(
                    "Loop found! Its bounding box parameters in fractional coordianates are: center (vertical %.3f, horizontal %.3f), height %.3f, width %.3f"
                    % (r, c, h, w)
                )
            else:
                self.log.debug("loop not found !")

            most_likely_click = description["most_likely_click"]
            v, h = most_likely_click
            self.log.debug(
                "Most likely click in fractional coordinates: (vertical %.3f, horizontal %.3f)"
                % (v, h)
            )
        else:
            self.log.debug("Loop not found. Click to the center.")

            v = 0.5
            h = 0.5

        # Save the debug results
        name_pattern = f"{os.getuid()}_{time.asctime().replace(' ', '_')}.jpg"
        directory = f"{os.getenv('HOME')}/murko"
        os.makedirs(directory, exist_ok=True)
        template = os.path.join(directory, name_pattern)
        plot_analysis([image_jpeg], analysis, image_paths=[template])

        return h, v

    def auto_ai_routine(self, zoom):

        self.goto_centring_phase()
        self.log.debug("Automatic 3 click centring")
        self.emit_progress_message("AI centring. Please wait...")

        # Define Zoom: temporary disable for now.
        # It will use whatewer zoom is selected.
        # zoom_hwobj = self.motor_hwobj_dict["zoom"]
        # zoom_hwobj.camera_hwobj.set_zoom(zoom)

        motor_pos = self.automatic_ai_centring()
        self.move_to_centred_position(motor_pos)
        return motor_pos

    def centring_done(self, centring_procedure, motor_p=None):
        """
        Descript. :
        """
        try:
            motor_pos = centring_procedure.get()
            if isinstance(motor_pos, gevent.GreenletExit):
                raise motor_pos
        except Exception:
            logging.exception("Could not complete centring")
            self.emit_centring_failed()
        else:
            self.emit_progress_message("Moving sample to centred position...")
            self.emit_centring_moving()

            try:
                logging.getLogger("HWR").debug(
                    "Centring finished. Moving motors to position %s" % str(motor_pos)
                )
                self.move_to_motors_positions(motor_pos, wait=True)
            except Exception:
                logging.exception("Could not move to centred position")
                self.emit_centring_failed()
            else:
                # if 3 click centring move -180. well. dont, in principle the calculated
                # centred positions include omega to initial position
                pass
                # if not self.in_plate_mode():
                #    logging.getLogger("HWR").debug("Centring finished. Moving omega back to initial position")
                #    self.motor_hwobj_dict['phi'].set_value_relative(-180, timeout=None)
                #    logging.getLogger("HWR").debug("         Moving omega done")

            if (
                self.current_centring_method
                == GenericDiffractometer.CENTRING_METHOD_AUTO
            ):
                self.emit("newAutomaticCentringPoint", motor_pos)
            self.ready_event.set()
            self.centring_time = time.time()
            self.emit_centring_successful()
            self.emit_progress_message("")

    def automatic_centring(self):
        """Automatic centring procedure"""

        # TODO: check the scaling change and reference image coordinates after zoom update.
        # check coordinates -move to the beam
        # self.auto_ai_routine(0)
        # self.auto_ai_routine(0.3)

        self.current_centring_procedure = gevent.spawn(self.auto_ai_routine, 0)
        self.current_centring_procedure.link(self.centring_done)

    def automatic_ai_centring(
        self,
        n_clicks=3,
        alignmenty_direction=1.0,
        alignmentz_direction=1.0,
        centringx_direction=1.0,
        centringy_direction=1.0,
    ):
        """Automatic centring procedure"""
        self.log.debug("** Autocentering with murko is started **")

        self.current_centring_method == GenericDiffractometer.CENTRING_METHOD_AUTO
        logging.getLogger("user_level_log").info("Starting AI centring")
        _start = time.time()
        result_position = {}

        reference_position = self.get_positions()

        vertical_clicks = []
        horizontal_clicks = []
        vertical_discplacements = []
        horizontal_displacements = []
        omegas = []
        images = []
        calibrations = []

        if isinstance(self.nclicks, int) and self.nclicks >= 3:
            n_clicks = self.nclicks

        logging.getLogger("user_level_log").info(
            "expected number of clicks %d" % (n_clicks)
        )

        if self.step is not None:
            step = self.step
        else:
            step = 360.0 / (n_clicks)

        logging.getLogger("user_level_log").info("Default centring step %.2f" % (step))

        for k in range(n_clicks):
            image = HWR.beamline.sample_view.camera.last_jpeg
            x_click, y_click = self.predict_click_pix(image)

            x = x_click * int(os.getenv("MURKO_SIZEX"))
            y = y_click * int(os.getenv("MURKO_SIZEY"))

            calibration = np.array(
                [1.0 / self.pixels_per_mm_y, 1.0 / self.pixels_per_mm_x]
            )
            omega = self.centring_phi.motor.get_value()

            vertical_clicks.append(y)
            horizontal_clicks.append(x)
            omegas.append(omega)
            images.append(image)
            calibrations.append([calibration])

            x -= self.beam_position[0]
            x /= self.pixels_per_mm_x
            y -= self.beam_position[1]
            y /= self.pixels_per_mm_y
            vertical_discplacements.append(y)
            horizontal_displacements.append(x)

            logging.getLogger("HWR").info("click %d %f %f %f" % (k + 1, omega, x, y))

            dev_gonio = DeviceProxy("p11/servomotor/eh.1.01")
            if k <= n_clicks:
                while str(dev_gonio.State()) != "ON":
                    time.sleep(0.1)
                self.centring_phi.set_value(omega + step, timeout=0)

                while str(dev_gonio.State()) != "ON":
                    time.sleep(0.1)

        vertical_discplacements = np.array(vertical_discplacements) * 1.0e3
        angles = np.radians(omegas)

        initial_parameters = lmfit.Parameters()
        initial_parameters.add_many(
            ("c", 0.0, True, -5e3, +5e3, None, None),
            ("r", 0.0, True, 0.0, 4e3, None, None),
            ("alpha", -np.pi / 3, True, -2 * np.pi, 2 * np.pi, None, None),
            ("front", 0.01, True, 0.0, 1.0, None, None),
            ("back", 0.005, True, 0.0, 1.0, None, None),
            ("n", 1.31, True, 1.29, 1.33, None, None),
            ("beta", 0.0, True, -2 * np.pi, +2 * np.pi, None, None),
        )

        fit_y = lmfit.minimize(
            self.refractive_model_residual,
            initial_parameters,
            method="nelder",
            args=(angles, vertical_discplacements),
        )
        optimal_params = fit_y.params
        v = optimal_params.valuesdict()
        c = v["c"]
        r = v["r"]
        alpha = v["alpha"]
        front = v["front"]
        back = v["back"]
        n = v["n"]
        beta = v["beta"]

        c *= 1.0e-3
        r *= 1.0e-3
        front *= 1.0e-3
        back *= 1.0e-3

        horizontal_center = np.mean(horizontal_displacements)

        d_sampx = centringx_direction * r * np.sin(alpha)
        d_sampy = centringy_direction * r * np.cos(alpha)
        d_y = alignmenty_direction * horizontal_center
        d_z = alignmentz_direction * c

        move_vector_dictionary = {
            "phiz": d_z,
            "phiy": d_y,
            "sampx": d_sampx,
            "sampy": d_sampy,
        }

        for motor in reference_position:
            result_position[motor] = reference_position[motor]
            if motor in move_vector_dictionary:
                result_position[motor] += move_vector_dictionary[motor]

        _end = time.time()
        duration = _end - _start
        self.log.info(
            "input and analysis in murko centring took %.3f seconds" % duration
        )

        results = {
            "vertical_clicks": vertical_clicks,
            "horizontal_clicks": horizontal_clicks,
            "vertical_discplacements": vertical_discplacements,
            "horizontal_displacements": horizontal_displacements,
            "omegas": omegas,
            "angles": angles,
            "calibrations": calibrations,
            "reference_position": reference_position,
            "result_position": result_position,
            "duration": duration,
            "vertical_optimal_parameters": v,
            "move_vector_dictionary": move_vector_dictionary,
        }

        name_pattern = "%s_%s" % (os.getuid(), time.asctime().replace(" ", "_"))
        directory = "%s/manual_optical_alignment" % os.getenv("HOME")

        template = os.path.join(directory, name_pattern)

        if not os.path.isdir(directory):
            os.makedirs(directory)

        clicks_filename = "%s_clicks.pickle" % template
        f = open(clicks_filename, "wb")
        pickle.dump(results, f)
        f.close()

        self.log.info("AI finished in %.3f seconds" % (time.time() - _start))

        result_position = {}
        result_position["phi"] = reference_position["phi"]
        for key in move_vector_dictionary:
            result_position[key] = (
                reference_position[key] + 1e3 * move_vector_dictionary[key]
            )

        self.log.info("result_position %s" % str(result_position))

        self.ai_finished = True

        return result_position

    def is_ready(self):
        """
        Descript. :
        """
        return True

    def is_valid(self):
        """
        Descript. :
        """
        return True

    def invalidate_centring(self):
        """
        Descript. :
        """
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            self.emit("centringInvalid", ())

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """
        Descript. :
        """
        return

    def refresh_omega_reference_position(self):
        """
        Descript. :
        """
        return

    def get_omega_axis_position(self):
        """
        Descript. :
        """
        return self.motor_hwobj_dict["phi"].get_value()

    def beam_position_changed(self, value):
        """
        Descript. :
        """
        self.beam_position = value

    def get_current_centring_method(self):
        """
        Descript. :
        """
        return self.current_centring_method

    def motor_positions_to_screen(self, centred_positions_dict):
        """
        Descript. :
        """
        self.update_zoom_calibration()

        beam_xc, beam_yc = self.beam_position
        phi_pos = self.motor_hwobj_dict["phi"].get_value()

        sampx_c = centred_positions_dict["sampx"]
        sampy_c = centred_positions_dict["sampy"]
        phiy_c = centred_positions_dict["phiy"]

        sampx_pos = self.centring_sampx.motor.get_value()
        sampy_pos = self.centring_sampy.motor.get_value()
        phiy_pos = self.centring_phiy.motor.get_value()

        sampx_d = sampx_c - sampx_pos
        sampy_d = sampy_c - sampy_pos
        phiy_d = phiy_c - phiy_pos

        # convert to mms
        sampx_d = self.centring_sampx.units_to_mm(sampx_d)
        sampy_d = self.centring_sampy.units_to_mm(sampy_d)
        phiy_d = self.centring_phiy.units_to_mm(phiy_d)

        cphi = math.cos(math.radians(phi_pos))
        sphi = math.sin(math.radians(phi_pos))

        dx = sampx_d * cphi - sampy_d * sphi
        dy = sampx_d * sphi + sampy_d * cphi

        xdist = phiy_d * self.pixels_per_mm_x
        ydist = dy * self.pixels_per_mm_y

        x = beam_xc + xdist
        y = beam_yc + ydist

        return x, y

    def start_auto_focus(self):
        """
        Descript. :
        """
        return

    def start_manual_centring(self, sample_info=None, wait_result=None):
        """
        """
        self.goto_centring_phase()
        self.log.debug(
            "Manual 3 click centring. using sample centring module: %s"
            % self.use_sample_centring
        )
        self.emit_progress_message("Manual 3 click centring...")

        self.current_centring_procedure = gevent.spawn(self.manual_centring)
        self.current_centring_procedure.link(self.centring_done)

    def circle_model(self, angles, c, r, alpha):
        return c + r * np.cos(angles - alpha)

    def circle_model_residual(self, parameters, angles, data):
        c, r, alpha = parameters
        model = self.circle_model(angles, c, r, alpha)
        return self.cost(data, model, normalize=True)

    def circle_model_residual2(self, parameters, angles, data):
        v = parameters.valuesdict()
        c = v["c"]
        r = v["r"]
        alpha = v["alpha"]
        model = self.circle_model(angles, c, r, alpha)

        return self.cost_array(data, model)

    def refractive_model(self, t, c, r, alpha, front, back, n, beta):
        return self.circle_model(t, c, r, alpha) - self.shift(t, front, back, n, beta)

    def refractive_model_residual(self, parameters, angles, data):
        v = parameters.valuesdict()
        c = v["c"]
        r = v["r"]
        alpha = v["alpha"]
        front = v["front"]
        back = v["back"]
        n = v["n"]
        beta = v["beta"]
        model = self.refractive_model(angles, c, r, alpha, front, back, n, beta)

        return self.cost_array(data, model)

    def cost(self, data, model, factor=1.0, normalize=False):
        if normalize == True:
            factor = 1.0 / (2 * len(model))
        return factor * np.sum(np.sum(np.abs(data - model) ** 2))

    def cost_array(self, data, model):
        return np.abs(data - model) ** 2

    def i(self, t, n):
        return np.arcsin(np.sin(t) / n)

    def planparallel_shift(self, depth, t, n, sense=1):
        i = self.i(t, n)
        return -depth * np.sin(sense * t - i) / np.cos(i)

    def shift(self, t, f, b, n, beta):
        t = t - beta
        dt = np.degrees(t)
        s = np.zeros(dt.shape)
        t_base = t % (2 * np.pi)
        mask = np.where(((t_base < 3 * np.pi / 2) & (t_base >= np.pi / 2)), 1, 0)
        s[mask == 0] = self.planparallel_shift(f, t_base[mask == 0], n, sense=1)
        s[mask == 1] = self.planparallel_shift(b, t_base[mask == 1], n, sense=-1)
        return s


    def manual_centring(self, phi_range=120, n_points=3):
        """
        Descript. :
        """
        X = []
        Y = []
        PHI = []

        beam_xc, beam_yc = self.beam_position
        self.log.debug("STARTING Manual Centring")

        motor_positions = {
                'phiy': self.centring_phiy.motor.get_value(),
                'sampx': self.centring_sampx.motor.get_value(),
                'sampy': self.centring_sampy.motor.get_value(),
                'phi': self.centring_phi.motor.get_value(),
                }

        phi_mot = self.centring_phi.motor
        phi_start_pos = phi_mot.get_value()

        for click in range(n_points):
            self.user_clicked_event = AsyncResult()
            x, y = self.user_clicked_event.get()
            if click < 2:
                phi_mot.set_value_relative(phi_range)

            X.append(x)
            Y.append(y)
            PHI.append(phi_mot.get_value())

        DX = []
        DY = []
        ANG = []

        P = []
        Q = []

        for i in range(n_points):
            dx = X[i] - beam_xc
            dy = Y[i] - beam_yc
            ang = math.radians( PHI[i])

            DX.append(dx)
            DY.append(dy)
            ANG.append(ang)

        for i in range(n_points):
            y0 = DY[i]
            ang0 = ANG[i]
            if i < (n_points-1):
                y1 = DY[i+1]
                ang1 = ANG[i+1]
            else:
                y1 = DY[0]
                ang1 = ANG[0]

            p = ( y0*math.sin(ang1) - y1*math.sin(ang0) ) / math.sin(ang1-ang0)
            q = ( y0*math.cos(ang1) - y1*math.cos(ang0) ) / math.sin(ang1-ang0)

            P.append(p)
            Q.append(q)

        x_s = -sum(Q)/n_points
        y_s = sum(P)/n_points
        z_s = sum(DX)/n_points

        x_d_mm = x_s / self.pixels_per_mm_y
        y_d_mm = y_s / self.pixels_per_mm_y
        z_d_mm = z_s / self.pixels_per_mm_x

        x_d = self.centring_sampx.mm_to_units(x_d_mm)
        y_d = self.centring_sampy.mm_to_units(y_d_mm)
        z_d = self.centring_phiy.mm_to_units(z_d_mm)

        sampx_mot = self.centring_sampx.motor
        sampy_mot = self.centring_sampy.motor
        phiy_mot = self.centring_phiy.motor

        x_pos = sampx_mot.get_value() + x_d
        y_pos = sampy_mot.get_value() + y_d
        z_pos = phiy_mot.get_value() + z_d

        motor_positions['phiy'] = z_pos
        motor_positions['sampx'] = x_pos
        motor_positions['sampy'] = y_pos
        return motor_positions

    def get_positions(self):
        sampx_pos = self.motor_hwobj_dict["sampx"].get_value()
        sampy_pos = self.motor_hwobj_dict["sampy"].get_value()
        phiy_pos = self.motor_hwobj_dict["phiy"].get_value()
        phiz_pos = self.motor_hwobj_dict["phiz"].get_value()
        phi_pos = self.motor_hwobj_dict["phi"].get_value()

        return {
            "phi": phi_pos,
            "phiy": phiy_pos,
            "phiz": phiz_pos,
            "sampx": sampx_pos,
            "sampy": sampy_pos,
        }

    def move_to_beam(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """

        # calculate distance from clicked position to center in mm
        dx = (x - self.beam_position[0]) / self.pixels_per_mm_x
        dy = (y - self.beam_position[1]) / self.pixels_per_mm_y

        phi = self.centring_phi.get_value()

        cphi = math.cos(math.radians(phi))
        sphi = math.sin(math.radians(phi))

        samp_y = dy * cphi
        samp_x = dy * sphi

        # convert to microns if necessary
        samp_x = self.centring_sampx.mm_to_units(samp_x)
        samp_y = self.centring_sampy.mm_to_units(samp_y)
        x_dist = self.centring_phiy.mm_to_units(dx)

        samp_x_pos = self.centring_sampx.motor.get_value() + samp_x
        samp_y_pos = self.centring_sampy.motor.get_value() + samp_y
        phiy = self.centring_phiy.motor.get_value() + x_dist

        self.centring_sampx.motor.set_value(samp_x_pos, timeout=0)
        self.centring_sampy.motor.set_value(samp_y_pos, timeout=0)
        self.centring_phiy.motor.set_value(phiy, timeout=0)

    def start_move_to_beam(self, coord_x=None, coord_y=None, omega=None):
        """
        Descript. :
        """
        self.centring_time = time.time()
        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centring_status = {
            "valid": True,
            "startTime": curr_time,
            "endTime": curr_time,
        }
        motors = self.get_positions()
        motors["beam_x"] = 0.1
        motors["beam_y"] = 0.1
        self.centring_status["motors"] = motors
        self.centring_status["valid"] = True
        self.centring_status["angleLimit"] = False
        self.emit_progress_message("")
        self.accept_centring()
        self.current_centring_method = None
        self.current_centring_procedure = None

    def re_emit_values(self):
        self.emit("zoomMotorPredefinedPositionChanged", None, None)
        omega_ref = [0, 238]
        self.emit("omegaReferenceChanged", omega_ref)

    def move_kappa_and_phi(self, kappa, kappa_phi):
        return

    def get_osc_max_speed(self):
        return 66

    def get_osc_limits(self):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_osc_dynamic_limits(self):
        """Returns dynamic limits of oscillation axis"""
        return (0, 20)

    def get_scan_dynamic_limits(self, speed=None):
        return (-360, 360)

    def stop_motion(self):
        self.acq_abort()
        self.acq_window_off_cmd()
        self.acq_off_cmd()

    def move_omega_relative(self, relative_angle):
        self.wait_omega_on()
        self.motor_hwobj_dict["phi"].set_value_relative(relative_angle, 5)

    def set_pso_control_arm(self, start_angle, stop_angle):
        """This Method sets the Tango Device Server "proxyGoniometer" into the PSOcontrolArm mode.
        For this mode, there is a  contact(pin) on controller, which will be set to High(Low)
        State, to trigger "start acquisition"("end acquisition") events of Detector-FastShutter.
        """
        try:
            if start_angle <= stop_angle:
                self.lower_bound_ch.set_value(start_angle)
                self.upper_bound_ch.set_value(stop_angle)
            else:
                self.lower_bound_ch.set_value(stop_angle)
                self.upper_bound_ch.set_value(start_angle)

            self.acq_arm_cmd()
        except:
            print(
                "++++++++++++++++++++++++++++++++++++++++++++++++ ERROR setting the PSO********************************************************"
            )

    def move_omega(self, angle):
        self.wait_omega_on()
        self.wait_omega()
        self.motor_hwobj_dict["phi"].set_value(angle, timeout=0)

    def move_omega_with_calibration(self, angle):
        self.wait_omega_on()

        # Code below is calibration so that angle is always stays within 360 degrees.
        # TODO: check the behaviour
        currentAngle = self.get_omega_position()

        if currentAngle < 0:
            currentAngle = abs(currentAngle)
            loadAngleNew = currentAngle % 360
            self.omega_calibrate(-loadAngleNew)
            self.motor_hwobj_dict["phi"].set_value(angle, timeout=0)
        else:
            loadAngleNew = currentAngle % 360
            self.omega_calibrate(loadAngleNew)
            self.motor_hwobj_dict["phi"].set_value(angle, timeout=0)

    def wait_omega_on(self, timeout=30):
        """
        Wait until the omega device finishes its movement and reaches the "ON" state.
        """
        dev_omega = DeviceProxy("p11/servomotor/eh.1.01")
        start_time = time.time()

        while True:
            try:
                current_state = str(dev_omega.State())
                if current_state == "ON":
                    logging.info("Omega device has finished movement and is now ON.")
                    break
            except DevFailed as e:
                logging.error(f"Error communicating with the omega device: {e}")

            if time.time() - start_time > timeout:
                logging.error(
                    "Timeout while waiting for omega device to finish movement and turn ON."
                )
                break

            time.sleep(0.1)

    def get_omega_position(self):
        return self.motor_hwobj_dict["phi"].get_value()

    def get_omega_velocity(self):
        return self.motor_hwobj_dict["phi"].get_velocity()

    def set_omega_velocity(self, value):
        return self.motor_hwobj_dict["phi"].set_velocity(value)

    def omega_calibrate(self, value):
        return self.motor_hwobj_dict["phi"].calibrate(value)

    def wait_omega(self):
        while self.motor_hwobj_dict["phi"].is_moving():
            time.sleep(0.05)

    def get_point_from_line(self, point_one, point_two, index, images_num):
        return point_one.as_dict()

    def set_phase(self, phase, timeout=None):
        self.set_phase_task = gevent.spawn(self.goto_phase, phase)

    def get_phase(self):
        return self.current_phase

    def motor_state_changed(self, state=None):

        new_state = DiffractometerState.Ready

        if self.phase_state == self.PHASE_STATES.MOVING:
            new_state = DiffractometerState.Moving
        else:
            for motname, motor in self.motor_hwobj_dict.items():
                mot_state = motor.get_state()
                if mot_state == HardwareObjectState.UNKNOWN:
                    new_state = DiffractometerState.Unknown
                    break
                elif mot_state == HardwareObjectState.FAULT:
                    new_state = DiffractometerState.Fault
                    break
                elif motor.is_moving():
                    new_state = DiffractometerState.Moving
                    break

        if new_state != self.diffractometer_state:
            self._update_state(new_state)

    def _update_state(self, new_state):
        self.emit("minidiffStateChanged", (new_state,))
        self.diffractometer_state = new_state

    @task
    def goto_phase(self, phase):
        self.log.debug("Starting phase change  - setting phase to %s\n" % phase)

        self.phase_state = self.PHASE_STATES.MOVING
        self.motor_state_changed()

        self.emit("minidiffPhaseStateChanged", (self.PHASE_STATES.MOVING,))

        if phase.lower() == "centring":
            self.goto_centring_phase(wait=True)
        elif phase.lower() == "transfer":
            self.goto_transfer_phase(wait=True)
        elif phase.lower() == "datacollection":
            self.goto_collect_phase(wait=True)
        elif phase.lower() == "beamlocation":
            self.goto_beam_phase(wait=True)
        else:
            self.user_log.debug("Unknown diffractometer phase: %s\n", phase)

        self.waiting_phase = True
        self.wait_phase()

    def wait_phase(self, timeout=140):
        start_wait = time.time()

        self.log.debug(" WAITING PHASE STARTED")
        while time.time() - start_wait < timeout:
            time.sleep(0.1)

            moving = False
            for ho in [
                self.detcover_hwobj,
                self.backlight_hwobj,
                self.beamstop_hwobj,
                self.collimator_hwobj,
                self.yag_hwobj,
                self.pinhole_hwobj,
            ]:

                if ho.is_moving():
                    moving = True

            if moving:
                continue

            break

        self.log.debug(" PHASE REACHED. NOW WAITING FOR OMEGA")
        self.wait_omega()

        gevent.sleep(0.6)  # allow for position events to arrive
        self.update_phase()
        self.motor_state_changed()

        # Extra waiting loop for the pinhole did not reached the top position because it is blocked.
        pinhole_states = ["200", "100", "50", "20", "down"]
        timeout = 140
        start_wait = time.time()
        self.log.debug(
            "================= Wait whiile pinholes are not blocked whille going up. Pinhole now in the position "
            + str(self.pinhole_hwobj.get_value())
        )
        while time.time() - start_wait < timeout:
            time.sleep(0.1)

            for st in pinhole_states:
                if (
                    self.pinhole_hwobj.get_value() == st
                    and not self.pinhole_hwobj.is_moving()
                ):
                    self.log.debug(
                        "Pinhole has reached position "
                        + str(self.pinhole_hwobj.get_value())
                    )
                    break
                else:
                    self.log.debug("Still waiting for the pinhole")
                    continue
            break

        self.log.debug(" PHASE CHANGED COMPLETED")
        self.waiting_phase = False

    def is_centring_phase(self):
        return self.get_phase() == GenericDiffractometer.PHASE_CENTRING

    def goto_centring_phase(self, wait=True):
        self.log.debug(" SETTING CENTRING PHASE ")

        self.phase_goingto = GenericDiffractometer.PHASE_CENTRING

        logging.getLogger("GUI").warning("Closing detector cover...")
        self.detcover_hwobj.close(timeout=0)

        logging.getLogger("GUI").warning("Setting backlight in...")
        self.backlight_hwobj.set_in()

        logging.getLogger("GUI").warning("Putting collimator down...")
        self.collimator_hwobj.set_value("down")

        logging.getLogger("GUI").warning("Setting collimator out...")
        self.beamstop_hwobj.set_value("out")

        logging.getLogger("GUI").warning("Moving yag down...")
        self.yag_hwobj.set_value("down")

        logging.getLogger("GUI").warning("Moving pinhole down...")
        self.pinhole_hwobj.set_value("down")

        if wait:
            self.wait_phase()

        logging.getLogger("GUI").info("Phase set is finished")

    def is_transfer_phase(self):
        return self.get_phase() == GenericDiffractometer.PHASE_TRANSFER

    def goto_transfer_phase(self, wait=True):

        self.log.debug(" SETTING TRANSFER PHASE ")
        self.phase_goingto = GenericDiffractometer.PHASE_TRANSFER
        self.moving_motors = True

        try:
            self.log.debug("  - close detector cover")
            self.detcover_hwobj.close(timeout=0)

            self.log.debug("  - setting backlight out")
            self.backlight_hwobj.set_out()

            self.log.debug("  - putting collimator down")
            self.collimator_hwobj.set_value("down")

            self.log.debug("  - setting beamstop out")
            self.beamstop_hwobj.set_value("out")

            self.log.debug("  - moving yag down")
            self.yag_hwobj.set_value("down")

            self.log.debug("  - moving pinhole down")

            if not self.ignore_pinhole:
                self.pinhole_hwobj.set_value("down")

            self.log.debug("  - moving omega to 0")

            self.move_omega(0)
            self.restore_position("transfer")

            self.log.debug("  - moving gonio tower to 0")
        finally:
            self.moving_motors = False
            self.update_phase()

        if wait:
            self.wait_phase()

        # sampx to 0
        # sampy to 0
        # microx, microy to 0

    def detector_cover_open(self, wait=True):
        self.detcover_hwobj.open(timeout=0)
        if wait:
            self.wait_detcover(state="close")

    def detector_cover_close(self, wait=True):
        self.detcover_hwobj.close(timeout=0)
        if wait:
            self.wait_detcover(state="close")

    def wait_detcover(self, state, timeout=60):
        start_time = time.time()
        while time.time() - start_time > timeout:
            if state == "open" and self.detcover_hwobj.is_open:
                break
            elif state == "close" and self.detcover_hwobj.is_closed:
                break
            gevent.sleep(0.5)

    def is_collect_phase(self):
        if self.get_phase() == GenericDiffractometer.PHASE_COLLECTION:
            return True
        else:
            self.wait_phase()

        return self.get_phase() == GenericDiffractometer.PHASE_COLLECTION

    def goto_collect_phase(self, wait=True):
        self.phase_goingto = GenericDiffractometer.PHASE_COLLECTION

        self.log.debug(" SETTING DATA COLLECTION PHASE ")

        logging.getLogger("GUI").warning("Moving pinhole out...")
        self.backlight_hwobj.set_out()

        logging.getLogger("GUI").warning("Moving collimator up...")
        self.collimator_hwobj.set_value("up")

        logging.getLogger("GUI").warning("Moving beamstop in...")
        self.beamstop_hwobj.set_value("in")

        logging.getLogger("GUI").warning("Moving yag down...")
        self.yag_hwobj.set_value("down")

        self.log.debug("=========  - checking pinhole ===============")

        # If the pinhole is Down set pinhole to 200
        if self.pinhole_hwobj.get_value() == "down":
            logging.getLogger("GUI").warning("Pinhole is down. Setting pinhole to 200.")
            self.pinhole_hwobj.set_value("200")

        if wait:
            self.wait_phase()

        logging.getLogger("GUI").info("Phase set is finished")

    def goto_beam_phase(self, wait=True):
        self.phase_goingto = GenericDiffractometer.PHASE_BEAM

        self.log.debug(" SETTING BEAM LOCATION PHASE ")

        self.log.debug("  - setting backlight out")
        self.backlight_hwobj.set_out()

        self.log.debug("  - setting beamstop in")
        self.beamstop_hwobj.set_value("in")
        self.log.debug("  - moving scintillator up")
        self.yag_hwobj.set_value("yag")

        if wait:
            self.wait_phase()

    def update_phase(self, value=None):

        omega_pos = self.get_omega_position()

        cover_open = self.detcover_hwobj.is_open
        cover_closed = self.detcover_hwobj.is_closed
        blight_in = self.backlight_hwobj.is_in()
        blight_out = self.backlight_hwobj.is_out()
        collim = self.collimator_hwobj.get_value()
        bstop = self.beamstop_hwobj.get_value()
        pinh = self.pinhole_hwobj.get_value()
        yag = self.yag_hwobj.get_value()


        omega_moving = self.omega_hwobj.is_moving()
        cover_moving = self.detcover_hwobj.is_moving()
        light_moving = self.backlight_hwobj.is_moving()

        current_phase = GenericDiffractometer.PHASE_UNKNOWN
        missing = []

        if self.phase_goingto is GenericDiffractometer.PHASE_CENTRING:
            if not blight_in:
                missing.append("lightin")
            if not cover_closed:
                missing.append("cover_closed")
            if not collim == "down":
                missing.append("collim_down")
            if not yag == "down":
                missing.append("yag_out")
            if not bstop == "out":
                missing.append("bstop_out")
            if not pinh == "down" and not self.ignore_pinhole:
                missing.append("pinh_down")

            if not missing:
                current_phase = GenericDiffractometer.PHASE_CENTRING

        elif self.phase_goingto is GenericDiffractometer.PHASE_TRANSFER:
            if not blight_out:
                missing.append("lightout")
            if not cover_closed:
                missing.append("cover_closed")
            if not collim == "down":
                missing.append("collim_down")
            if not bstop == "out":
                missing.append("bstop_out")
            if not yag == "down":
                missing.append("yag_out")
            if not pinh == "down" and not self.ignore_pinhole:
                missing.append("pinh_down")
            if abs(omega_pos) >= 0.01:
                missing.append("omega_zero")

            if self.moving_motors:
                missing.append("motors_done")

            if not missing:
                self.log.debug("going to transfer done")
                current_phase = GenericDiffractometer.PHASE_TRANSFER
            else:
                self.log.debug("going to transfer. missing %s" % str(missing))

        elif self.phase_goingto is GenericDiffractometer.PHASE_COLLECTION:
            if not blight_out:
                missing.append("lightout")
            # if not cover_open:
            #    missing.append("cover_opened")
            if not collim == "up":
                missing.append("collim_up")
            if not bstop == "in":
                missing.append("bstop_in")
            if not yag == "down":
                missing.append("yag_out")

            if not missing:
                current_phase = GenericDiffractometer.PHASE_COLLECTION
        else:
            current_phase = GenericDiffractometer.PHASE_UNKNOWN

        # if blight_in and cover_closed and \
        # collim == "Down" and bstop == "Out" and \
        # yag == "Out" and pinh == "Down":
        # current_phase = GenericDiffractometer.PHASE_CENTRING
        # elif blight_out and cover_closed and \
        # collim == "Down" and bstop == "Out" and \
        # yag == "Out" and pinh == "Down" and \
        # abs(omega_pos) < 0.01:
        # current_phase = GenericDiffractometer.PHASE_TRANSFER
        # elif blight_out and cover_open and \
        # collim == "Up" and bstop == "In" and \
        # yag == "Out":
        # current_phase = GenericDiffractometer.PHASE_COLLECTION

        if self.phase_goingto == current_phase:
            self.log.debug("PHASE REACHED - %s" % self.phase_goingto)

            self.phase_goingto = None

        if current_phase == GenericDiffractometer.PHASE_UNKNOWN:
            if self.phase_goingto is not None:
                self.log.debug("PHASE (%s) NOT REACHED YET" % str(self.phase_goingto))
                self.log.debug("  waiting for: " + ",".join(missing))
        else:
            if self.current_phase != current_phase:
                self.log.debug("PHASE changed to %s" % current_phase)
                self.current_phase = current_phase
                self.emit("minidiffPhaseChanged", (self.current_phase,))

        # if omega_moving or cover_moving or light_moving:
        #    phase_state = self.PHASE_STATES.MOVING
        # else:
        # phase_state = self.PHASE_STATES.READY

        if self.phase_goingto:
            phase_state = self.PHASE_STATES.MOVING
        else:
            phase_state = self.PHASE_STATES.READY

        if phase_state != self.phase_state:
            self.emit("minidiffPhaseStateChanged", (phase_state,))
            self.phase_state = phase_state
            self.motor_state_changed()

    def save_position(self, position_name):
        saved_position = {}
        for motname in self.save_motor_list:
            saved_position[motname] = self.motor_hwobj_dict[motname].get_value()
        saved_position["pinhole"] = self.pinhole_hwobj.get_value()
        saved_position["backlight"] = self.backlight_hwobj.get_value()
        self._saved_position[position_name] = saved_position
        self.log.debug("P11NanoDiff - saving positions for %s" % position_name)
        for name, value in saved_position.items():
            self.log.debug("     %s - %s  " % (name, value))

    def wait_position_ready(self, timeout=70):
        t0 = time.time()

        while (time.time() - t0) < timeout:
            busy = False
            for motname in self.save_motor_list:
                if not self.motor_hwobj_dict[motname].is_ready():
                    busy = True
                    state = self.motor_hwobj_dict[motname].get_state()
                    self.log.debug(
                        "  - motor %s is not ready. it is %s" % (motname, str(state))
                    )

            time.sleep(0.2)

            if not busy:
                break

        self.log.debug("P11NanoDiff -  motors ready ")

        while (time.time() - t0) < timeout:
            busy = False
            if self.pinhole_hwobj.is_moving():
                busy = True
                self.log.debug(" - pinhole is not ready")
            if self.backlight_hwobj.is_moving():
                busy = True
                self.log.debug(" - backlight is not ready")

            if not busy:
                break
            time.sleep(0.2)

        if busy:
            self.log.error("Timeout waiting for motors to finish movement")

    def restore_position(self, position_name):
        self.log.debug("Restoring position for %s" % position_name)
        self.log.debug(" (available are: %s)" % self._saved_position.keys())
        positions = self._saved_position.get(position_name, None)

        if positions:
            for motname, position in positions.items():
                if motname not in ["pinhole", "backlight"]:
                    self.motor_hwobj_dict[motname].set_value(position)

            if "pinhole" in self._saved_position[position_name]:
                pinh_pos = self._saved_position[position_name]["pinhole"]
                if not self.ignore_pinhole:
                    self.pinhole_hwobj.set_position(pinh_pos)
            if "backlight" in self._saved_position[position_name]:
                light_value = self._saved_position[position_name]["backlight"]
                self.backlight_hwobj.set_value(light_value)

            self.wait_position_ready()
        else:
            self.log.error("No transfer positions saved for %s" % position_name)

        self.update_phase()

    # Helper function to move individual motors
    def move_motor(self, motor_name, position, motor_dict, timeout):
        if motor_name in motor_dict:
            self.log.debug(f"Moving motor '{motor_name}' to position {position}")
            motor_hwobj = self.motor_hwobj_dict.get(motor_name)
            if motor_hwobj and position is not None:
                motor_hwobj.set_value(position, timeout=None)
                gevent.sleep(0.1)
                self.wait_device_ready(timeout)
                self.log.debug(f"'{motor_name}' movement DONE")

    def move_motors(self, motor_positions, timeout=10):
        """
        Moves diffractometer motors to the requested positions.

        """

        # Check if motor_positions is None
        if motor_positions is None:
            self.log.debug("motor_positions is None")

        # Convert to dictionary if not already one
        if not isinstance(motor_positions, dict):
            try:
                motor_positions = motor_positions.as_dict()
            except RuntimeWarning:
                self.log.rror(
                    "motor_positions is not a dict and cannot be converted using as_dict()"
                )
                return

        # Move translation motors
        for motor in ["phiy", "phiz"]:
            if motor in motor_positions:
                self.move_motor(motor, motor_positions[motor], motor_positions, timeout)

        # Move alignment motors
        for motor in ["sampx", "sampy"]:
            if motor in motor_positions:
                self.move_motor(motor, motor_positions[motor], motor_positions, timeout)

        # Move 'phi' motor separately
        if "phi" in motor_positions:
            phi_position = motor_positions["phi"]
            self.log.debug(f"Finally moving motor 'phi' to position {phi_position}")
            motor_hwobj = self.motor_hwobj_dict.get("phi")
            if (
                motor_hwobj
                and phi_position is not None
                and abs(motor_hwobj.get_value() - phi_position) > 1.0e-4
            ):
                motor_hwobj.set_value(phi_position, timeout=None)
                gevent.sleep(0.1)
                self.log.debug("Phi movement DONE")

        # Move any remaining motors
        for motor, position in motor_positions.items():
            if motor not in ["phiy", "phiz", "phi", "sampx", "sampy"]:
                self.move_motor(motor, motor_positions[motor], motor_positions, timeout)

        # Delay polling for state in the
        # case of controller not reporting MOVING inmediately after cmd
        if self.delay_state_polling and self.delay_state_polling > 0:
            gevent.sleep(self.delay_state_polling)

        # Final check to ensure all devices are ready
        self.wait_device_ready(timeout)



from mxcubecore.BaseHardwareObjects import HardwareObject

"""
Move the beamstop or the capillary, using the exporter protocol
Example xml file:
<object class="MicrodiffBeamstop">
  <username>Beamstop</username>
  <object role="beamstop" hwrid="/udiff_beamstop"></object>
  <save_cmd_name>saveBeamstopBeamPosition</save_cmd_name>
  <motors>
    <object role="horizontal" hwrid="/bstopy"></object>
    <object role="vertical" hwrid="/bstopz"></object>
  </motors>
</object>

Example udiff_beamstop.xml
<object class="MicrodiffInOut">
  <username>Beamstop</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <cmd_name>BeamstopPosition</cmd_name>
  <private_state>{"OFF":"out", "BEAM":"in"}</private_state>
  <timeout>100</timeout>
</object>

Example bstopy.xml (for bstopz only the motor name changes)
<object class="MD2Motor">
  <username>bstopy</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <actuator_name>BeamstopY</actuator_name>
  <GUIstep>0.01</GUIstep>
   <unit>1e-3</unit>
</object>

When used with capillary, only the command and motor names change.
Example capillary xml file:
<object class="MicrodiffBeamstop">
  <username>Capillary</username>
  <object role="beamstop" hwrid="/udiff_capillary"></object>
  <save_cmd_name>saveCapillaryBeamPosition</save_cmd_name>
  <motors>
    <object role="horizontal" hwrid="/capy"></object>
    <object role="vertical" hwrid="/capz"></object>
  </motors>
</object>
"""


class MicrodiffBeamstop(HardwareObject):
    def init(self):
        self.beamstop = self.get_object_by_role("beamstop")
        self.beamstop.state_attr.connect_signal("update", self.checkPosition)

        self.motors = self["motors"]
        self.roles = self.motors.get_roles()

        save_cmd_name = self.get_property("save_cmd_name")
        self.beamstopSetInPosition = self.beamstop.add_command(
            {
                "type": "exporter",
                "name": "bs_set_in",
                "address": self.beamstop.get_property("exporter_address"),
            },
            save_cmd_name,
        )

        # next two lines - just to make the beamstop brick happy
        self.amplitude = 0
        self.positions = self.beamstop.moves

    def moveToPosition(self, name):
        if name == "in":
            self.beamstop.actuatorIn(wait=True)
        elif name == "out":
            self.beamstop.actuatorOut(wait=True)
        self.checkPosition()

    def connect_notify(self, signal):
        self.checkPosition()

    def is_ready(self):
        return True

    def get_state(self):
        return "READY"

    def get_value(self):
        return self.checkPosition(noEmit=True)

    def checkPosition(self, pos=None, noEmit=False):
        if pos is None:
            pos = self.beamstop.get_actuator_state()
        try:
            pos = self.beamstop.states[pos]
        except Exception:
            pass

        if not noEmit:
            if pos:
                self.emit("positionReached", pos)
            else:
                self.emit("noPosition", ())
        return pos

    def setNewPositions(self, name, newPositions):
        if name == "in":
            self.beamstopSetInPosition()

    def getRoles(self):
        return self.roles



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


from XSDataCommon import (
    XSDataBoolean,
    XSDataDouble,
    XSDataInteger,
    XSDataString,
)
from XSDataControlDozorv1_1 import XSDataInputControlDozor

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractOnlineProcessing import (
    AbstractOnlineProcessing,
)

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"


class DozorOnlineProcessing(AbstractOnlineProcessing):
    def create_processing_input_file(self, processing_input_filename):
        """Creates dozor input file base on data collection parameters

        :param processing_input_filename
        :type : str
        """
        input_file = XSDataInputControlDozor()
        input_file.setTemplate(XSDataString(self.params_dict["template"]))
        input_file.setFirst_image_number(
            XSDataInteger(self.params_dict["first_image_num"])
        )
        input_file.setLast_image_number(XSDataInteger(self.params_dict["images_num"]))
        input_file.setFirst_run_number(XSDataInteger(self.params_dict["run_number"]))
        input_file.setLast_run_number(XSDataInteger(self.params_dict["run_number"]))
        input_file.setLine_number_of(XSDataInteger(self.params_dict["lines_num"]))
        input_file.setReversing_rotation(
            XSDataBoolean(self.params_dict["reversing_rotation"])
        )
        pixel_size = HWR.beamline.detector.get_pixel_size()
        input_file.setPixelMin(XSDataInteger(pixel_size[0]))
        input_file.setPixelMax(XSDataInteger(pixel_size[1]))
        input_file.setBeamstopSize(XSDataDouble(self.beamstop_hwobj.get_size()))
        input_file.setBeamstopDistance(XSDataDouble(self.beamstop_hwobj.get_distance()))
        input_file.setBeamstopDirection(
            XSDataString(self.beamstop_hwobj.get_direction())
        )

        input_file.exportToFile(processing_input_filename)

    def batch_processed(self, batch):
        """Method called from EDNA via xmlrpc to set results

        :param batch: list of dictionaries describing processing results
        :type batch: lis
        """

        if self.started:
            for image in batch:
                self.results_raw["spots_num"][image[0] - 1] = image[1]
                self.results_raw["spots_resolution"][image[0] - 1] = image[3]
                self.results_raw["score"][image[0] - 1] = image[4]

            self.align_processing_results(batch[0][0] - 1, batch[-1][0] - 1)
            self.emit("processingResultsUpdate", False)



import logging
import math
import os
import sys
from datetime import datetime

from ALBAClusterJob import ALBAEdnaProcJob
from PyTango import DeviceProxy
from xaloc import XalocJob
from XSDataAutoprocv1_0 import XSDataAutoprocInput
from XSDataCommon import (
    XSDataFile,
    XSDataInteger,
    XSDataString,
)

from mxcubecore.BaseHardwareObjects import HardwareObject

sys.path.append("/beamlines/bl13/controls/devel/pycharm/ALBAClusterClient")


root = os.environ["POST_PROCESSING_SCRIPTS_ROOT"]
sls_script = os.path.join(root, "gphl/autoproc/autoproc.process.sl")


class ALBAAutoProcessing(HardwareObject):
    def init(self):
        HardwareObject.init(self)

        self.template_dir = self.get_property("template_dir")
        var_dsname = self.get_property("variables_ds")
        logging.getLogger("HWR").debug(
            "ALBAAutoProcessing INIT: var_ds=%s, template_dir=%s"
            % (var_dsname, self.template_dir)
        )
        self.var_ds = DeviceProxy(var_dsname)

    # input files for standard collection auto processing
    def create_input_files(self, xds_dir, mosflm_dir, dc_pars):
        fileinfo = dc_pars["fileinfo"]
        osc_seq = dc_pars["oscillation_sequence"][0]

        prefix = fileinfo["prefix"]
        runno = fileinfo["run_number"]

        exp_time = osc_seq["exposure_time"]

        start_angle = osc_seq["start"]
        nb_images = osc_seq["number_of_images"]
        start_img_num = osc_seq["start_image_number"]
        angle_increment = osc_seq["range"]

        wavelength = osc_seq.get("wavelength", 0)

        xds_template_name = "XDS_TEMPLATE.INP"
        mosflm_template_name = "mosflm_template.dat"

        xds_template_path = os.path.join(self.template_dir, xds_template_name)
        mosflm_template_path = os.path.join(self.template_dir, mosflm_template_name)

        xds_file = os.path.join(xds_dir, "XDS.INP")
        mosflm_file = os.path.join(mosflm_dir, "mosflm.dat")

        t = datetime.now()

        # PREPARE VARIABLES
        detsamdis = self.var_ds.detsamdis
        beamx, beamy = self.var_ds.beamx, self.var_ds.beamy

        mbeamx, mbeamy = beamy * 0.172, beamx * 0.172

        datarangestartnum = start_img_num
        datarangefinishnum = start_img_num + nb_images - 1
        backgroundrangestartnum = start_img_num
        spotrangestartnum = start_img_num
        if angle_increment != 0:
            minimumrange = int(round(20 / angle_increment))
        elif angle_increment == 0:
            minimumrange = 1
        if nb_images >= minimumrange:
            backgroundrangefinishnum = start_img_num + minimumrange - 1
        if nb_images >= minimumrange:
            spotrangefinishnum = start_img_num + minimumrange - 1
        if nb_images < minimumrange:
            backgroundrangefinishnum = start_img_num + nb_images - 1
        if nb_images < minimumrange:
            spotrangefinishnum = start_img_num + nb_images - 1

        testlowres = 8.0
        largestvector = (
            0.172
            * ((max(beamx, 2463 - beamx)) ** 2 + (max(beamy, 2527 - beamy)) ** 2) ** 0.5
        )
        testhighres = round(
            wavelength / (2 * math.sin(0.5 * math.atan(largestvector / detsamdis))), 2
        )
        lowres = 50.0
        highres = testhighres
        datafilename = prefix + "_" + str(runno) + "_????"
        mdatafilename = prefix + "_" + str(runno) + "_####.cbf"
        seconds = 5 * exp_time

        if angle_increment < 1 and not angle_increment == 0:
            seconds = 5 * exp_time / angle_increment

        # DEFINE SG/UNIT CELL
        spacegroupnumber = ""
        unitcellconstants = ""

        datapath_dir = os.path.abspath(xds_file).replace("PROCESS_DATA", "RAW_DATA")
        datapath_dir = os.path.dirname(os.path.dirname(datapath_dir)) + os.path.sep

        # CREATE XDS.INP FILE
        xds_templ = open(xds_template_path, "r").read()

        xds_templ = xds_templ.replace("###BEAMX###", str(round(beamx, 2)))
        xds_templ = xds_templ.replace("###BEAMY###", str(round(beamy, 2)))
        xds_templ = xds_templ.replace("###DETSAMDIS###", str(round(detsamdis, 2)))
        xds_templ = xds_templ.replace("###ANGLEINCREMENT###", str(angle_increment))
        xds_templ = xds_templ.replace("###WAVELENGTH###", str(wavelength))
        xds_templ = xds_templ.replace("###DATARANGESTARTNUM###", str(datarangestartnum))
        xds_templ = xds_templ.replace(
            "###DATARANGEFINISHNUM###", str(datarangefinishnum)
        )
        xds_templ = xds_templ.replace(
            "###BACKGROUNDRANGESTART###", str(backgroundrangestartnum)
        )
        xds_templ = xds_templ.replace(
            "###BACKGROUNDRANGEFINISHNUM###", str(backgroundrangefinishnum)
        )
        xds_templ = xds_templ.replace("###SPOTRANGESTARTNUM###", str(spotrangestartnum))
        xds_templ = xds_templ.replace(
            "###SPOTRANGEFINISHNUM###", str(spotrangefinishnum)
        )
        xds_templ = xds_templ.replace("###TESTLOWRES###", str(testlowres))
        xds_templ = xds_templ.replace("###TESTHIGHRES###", str(testhighres))
        xds_templ = xds_templ.replace("###LOWRES###", str(lowres))
        xds_templ = xds_templ.replace("###HIGHRES###", str(highres))
        xds_templ = xds_templ.replace("###DIRECTORY###", str(datapath_dir))
        xds_templ = xds_templ.replace("###FILENAME###", str(datafilename))
        xds_templ = xds_templ.replace("###SECONDS###", str(int(seconds)))
        xds_templ = xds_templ.replace(
            "###LYSOZYME_SPACE_GROUP_NUMBER###", str(spacegroupnumber)
        )
        xds_templ = xds_templ.replace(
            "###LYSOZYME_UNIT_CELL_CONSTANTS###", str(unitcellconstants)
        )

        open(xds_file, "w").write(xds_templ)

        # CREATE MOSFLM.DAT FILE

        mosflm_templ = open(mosflm_template_path, "r").read()

        mosflm_templ = mosflm_templ.replace("###DETSAMDIS###", str(round(detsamdis, 2)))
        mosflm_templ = mosflm_templ.replace("###BEAMX###", str(round(mbeamx, 2)))
        mosflm_templ = mosflm_templ.replace("###BEAMY###", str(round(mbeamy, 2)))
        mosflm_templ = mosflm_templ.replace("###DIRECTORY###", str(datapath_dir))
        mosflm_templ = mosflm_templ.replace("###FILENAME###", str(mdatafilename))
        mosflm_templ = mosflm_templ.replace("###WAVELENGTH###", str(wavelength))
        mosflm_templ = mosflm_templ.replace(
            "###DATARANGESTARTNUM###", str(datarangestartnum)
        )

        open(mosflm_file, "w").write(mosflm_templ)

        # CREATE EDNAPROC XML FILE
        collection_id = dc_pars["collection_id"]
        output_dir = dc_pars["ednaproc_dir"]

        ednaproc_input_file = os.path.join(
            output_dir, "EDNAprocInput_%d.xml" % collection_id
        )

        ednaproc_input = XSDataAutoprocInput()

        input_file = XSDataFile()
        path = XSDataString()
        path.set_value(xds_file)
        input_file.setPath(path)

        ednaproc_input.setInput_file(input_file)
        ednaproc_input.setData_collection_id(XSDataInteger(collection_id))

        # output_dir = XSDataFile()
        # outpath = XSDataString()
        # outpath.set_value(output_dir)
        # output_dir.setPath(path)

        # ednaproc_input.setOutput_directory( output_dir )

        ednaproc_input.exportToFile(ednaproc_input_file)

        self.input_file = ednaproc_input_file

    # trigger auto processing for standard collection
    def trigger_auto_processing(self, dc_pars):
        logging.getLogger("HWR").debug(
            " ALBAAutoProcessing. triggering auto processing."
        )

        dc_id = dc_pars["collection_id"]
        output_dir = dc_pars["ednaproc_dir"]

        logging.getLogger("HWR").debug("    - collection_id = %s " % dc_id)
        logging.getLogger("HWR").debug("    - output_dir    = %s " % output_dir)

        job = ALBAEdnaProcJob()
        input_file = self.input_file  # TODO

        job.run(dc_id, input_file, output_dir)


def test_hwo(hwo):
    ofile = "/tmp/edna/edna_result"
    odir = "/tmp/edna"
    test_input_file = "/beamlines/bl13/projects/cycle2018-I/2018012551-bcalisto/mx2018012551/DATA/20180131/PROCESS_DATA/characterisation_ref-Thrombin-TB-TTI1_A_run1_1/EDNAInput_2004391.xml"
    result = hwo.run_edna(test_input_file, ofile, odir)
    print(result)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

from mxcubecore.BaseHardwareObjects import HardwareObject


class ValueStateChannel(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.yellow_states = self.green_states = self.red_states = None
        self.yellow_limits = self.red_limits = None

        self.current_value = -1
        self.current_state = "ON"

        self.chan_value = self.get_channel_object("value")
        if self.chan_value is not None:
            self.chan_value.connect_signal("update", self.value_changed)

        self.chan_state = self.get_channel_object("state")
        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self.state_changed)

        self.update_delta = self.get_property("delta")

        green_states = self.get_property("ok_states")
        yellow_states = self.get_property("warning_states")
        red_states = self.get_property("alarm_states")

        yellow_limits = self.get_property("warning_limits")
        red_limits = self.get_property("alarm_limits")

        if green_states:
            self.green_states = [
                stat.strip().upper() for stat in green_states.split(",")
            ]
        if yellow_states:
            self.yellow_states = [
                stat.strip().upper() for stat in yellow_states.split(",")
            ]
        if red_states:
            self.red_states = [stat.strip().upper() for stat in red_states.split(",")]

        if yellow_limits:
            lims = [float(lim.strip()) for lim in yellow_limits.split(",")]
            if len(lims) == 2:
                if lims[0] > lims[1]:
                    self.yellow_limits = lims[1], lims[0]
                else:
                    self.yellow_limits = lims
        if red_limits:
            lims = [float(lim.strip()) for lim in red_limits.split(",")]
            if len(lims) == 2:
                if lims[0] > lims[1]:
                    self.red_limits = lims[1], lims[0]
                else:
                    self.red_limits = lims

    def is_ready(self):
        return self._state == self.STATES.READY

    def state_changed(self, state=None):
        if state is None:
            state = self.chan_state.get_value()

        self._state = str(state).upper()

        self.update_status()

    def value_changed(self, value=None):
        """
        Event called when value has been changed
        :param pos: float
        :return:
        """
        if value is None:
            value = self.chan_value.get_value()

        if value != self.current_value:
            self.current_value = value

            self.emit("valueChanged", (value))
            self.update_status()

    def update_status(self):
        _state = "ON"
        val = self.current_value

        if self.red_limits:
            llim, hlim = self.red_limits
            if llim > val or hlim < val:
                _state = "ALARM"

        if _state != "ALARM" and self.yellow_limits:
            llim, hlim = self.yellow_limits
            if llim > val or hlim < val:
                _state = "WARNING"

        if self._red_states and self._state in self._red_states:
            _state = "ALARM"
        elif self._yellow_states and self._state in self._yellow_states:
            _state = _state == "ALARM" and _state or "WARNING"

        if _state != self.current_state:
            self.current_state = _state
            self.emit("stateChanged", _state)



from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class SampleStage(HardwareObject):
    def init(self):
        self.__axis = []
        for token in self:
            if token.name() == "axis":
                axis_name = token.get_property("objectName")
                if axis_name is not None:
                    axis = HWR.get_hardware_repository().get_hardware_object(axis_name)
                    if axis is not None:
                        self.__axis.append(axis)

    def isSampleStage(self):
        return True

    def getAxisList(self):
        return self.__axis



import sys
import types

# LNLS
try:
    from types import InstanceType
except ImportError:
    InstanceType = object
import collections
import logging

import gevent


class cleanup:
    def __init__(self, *args, **keys):
        self.cleanup_funcs = args
        self.keys = keys

    def __enter__(self):
        pass

    def __exit__(self, *args):
        if self.cleanup_funcs:
            for cleanup_func in self.cleanup_funcs:
                if not isinstance(cleanup_func, collections.abc.Callable):
                    continue
                try:
                    cleanup_func(**self.keys)
                except Exception:
                    logging.exception(
                        "Exception while calling cleanup callback %s", cleanup_func
                    )
                    continue


class error_cleanup:
    def __init__(self, *args, **keys):
        self.error_funcs = args
        self.keys = keys

    def __enter__(self):
        pass

    def __exit__(self, *args):
        if args[0] is not None and self.error_funcs:
            logging.debug("Doing error cleanup")
            for error_func in self.error_funcs:
                if not isinstance(error_func, collections.abc.Callable):
                    continue
                try:
                    error_func(**self.keys)
                except Exception:
                    logging.exception(
                        "Exception while calling error cleanup callback %s", error_func
                    )
                    continue


class TaskException:
    def __init__(self, exception, error_string, tb):
        self.exception = exception
        self.error_string = error_string
        self.tb = tb


class wrap_errors(object):
    def __init__(self, func):
        """Make a new function from `func', such that it catches all exceptions
        and return it as a TaskException object
        """
        self.func = func

    def __call__(self, *args, **kwargs):
        func = self.func
        try:
            return func(*args, **kwargs)
        except Exception:
            sys.excepthook(*sys.exc_info())
            return TaskException(*sys.exc_info())

    def __str__(self):
        return str(self.func)

    def __repr__(self):
        return repr(self.func)

    def __getattr__(self, item):
        return getattr(self.func, item)


def task(func):
    def start_task(*args, **kwargs):
        if args and isinstance(args[0], InstanceType):
            logging.debug("Starting %s%s", func.__name__, args[1:])
        else:
            logging.debug("Starting %s%s", func.__name__, args)

        try:
            wait = kwargs["wait"]
        except KeyError:
            wait = True
        else:
            del kwargs["wait"]
        try:
            timeout = kwargs["timeout"]
        except KeyError:
            timeout = None
        else:
            del kwargs["timeout"]

        try:
            t = gevent.spawn(wrap_errors(func), *args, **kwargs)

            if wait:
                ret = t.get(timeout=timeout)
                if isinstance(ret, TaskException):
                    sys.excepthook(ret.exception, ret.error_string, ret.tb)
                    raise ret.exception(ret.error_string)
                else:
                    return ret
            else:
                t._get = t.get

                def special_get(self, *args, **kwargs):
                    ret = self._get(*args, **kwargs)
                    if isinstance(ret, TaskException):
                        sys.excepthook(ret.exception, ret.error_string, ret.tb)
                        raise ret.exception(ret.error_string)
                    else:
                        return ret

                setattr(t, "get", types.MethodType(special_get, t))

                return t
        except Exception:
            t.kill()
            raise

    return start_task



from mxcubecore.HardwareObjects import GrobMotor


class GrobGonio(GrobMotor.GrobMotor):
    def __init__(self, name):
        GrobMotor.GrobMotor.__init__(self, name)

    def oscillation(self, *args):
        self.motor.oscillation(*args)



import logging
import math
import time

import numpy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import Microdiff
from mxcubecore.HardwareObjects.sample_centring import CentringMotor


class MD3UP(Microdiff.Microdiff):
    def __init__(self, *args, **kwargs):
        Microdiff.Microdiff.__init__(self, *args, **kwargs)

    def init(self):
        Microdiff.Microdiff.init(self)
        phiy_ref = self["centringReferencePosition"].get_property("phiy")

        self.centringPhi = CentringMotor(self.phiMotor, direction=1)
        self.centringPhiz = CentringMotor(self.phizMotor)
        self.centringPhiy = CentringMotor(
            self.phiyMotor, direction=-1, reference_position=phiy_ref
        )
        self.centringSamplex = CentringMotor(self.sampleXMotor, direction=-1)
        self.centringSampley = CentringMotor(self.sampleYMotor)
        self.scan_nb_frames = 1

        # Raster scan attributes
        self.nb_frames = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "nbframes",
            },
            "ScanNumberOfFrames",
        )
        self.scan_range = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "scan_range",
            },
            "ScanRange",
        )
        self.scan_exposure_time = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "exposure_time",
            },
            "ScanExposureTime",
        )
        self.scan_start_angle = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_angle",
            },
            "ScanStartAngle",
        )
        self.scan_detector_gate_pulse_enabled = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "detector_gate_pulse_enabled",
            },
            "DetectorGatePulseEnabled",
        )
        self.scan_detector_gate_pulse_readout_time = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "detector_gate_pulse_readout_time",
            },
            "DetectorGatePulseReadoutTime",
        )

    def set_rotation_axis_position(self, value: float, motor_name="phiy"):
        self._set_rotation_axis_position(value, motor_name="phiy")

    # def getBeamPosX(self):
    #     return self.beam_info.get_beam_position_on_screen()[0]
    #
    # def getBeamPosY(self):
    #     return self.beam_info.get_beam_position_on_screen()[1]

    def setNbImages(self, number_of_images):
        self.scan_nb_frames = number_of_images

    def oscilScan(self, start, end, exptime, number_of_images, wait=False):
        if self.in_plate_mode():
            scan_speed = math.fabs(end - start) / exptime
            low_lim, hi_lim = map(float, self.scanLimits(scan_speed))
            if start < low_lim:
                raise ValueError("Scan start below the allowed value %f" % low_lim)
            elif end > hi_lim:
                raise ValueError("Scan end abobe the allowed value %f" % hi_lim)

        dead_time = HWR.beamline.detector.get_deadtime()

        self.scan_detector_gate_pulse_enabled.set_value(True)
        self.scan_detector_gate_pulse_readout_time.set_value(dead_time * 1000)

        if self.get_property("md_set_number_of_frames", False):
            self.nb_frames.set_value(number_of_images)
        else:
            self.nb_frames.set_value(1)

        scan_params = "1\t%0.3f\t%0.3f\t%0.4f\t1" % (start, (end - start), exptime)
        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_scan",
            },
            "startScanEx",
        )
        scan(scan_params)
        print("oscil scan started at ----------->", time.time())
        if wait:
            self._wait_ready(20 * 60)  # Timeout of 20 min
            print("finished at ---------->", time.time())

    def oscilScan4d(
        self, start, end, exptime, number_of_images, motors_pos, wait=False
    ):
        if self.in_plate_mode():
            scan_speed = math.fabs(end - start) / exptime
            low_lim, hi_lim = map(float, self.scanLimits(scan_speed))
            if start < low_lim:
                raise ValueError("Scan start below the allowed value %f" % low_lim)
            elif end > hi_lim:
                raise ValueError("Scan end abobe the allowed value %f" % hi_lim)

        if self.get_property("md_set_number_of_frames", False):
            self.nb_frames.set_value(number_of_images)
        else:
            self.nb_frames.set_value(1)

        params = "%0.3f\t%0.3f\t%f\t" % (start, (end - start), exptime)
        params += "%0.3f\t" % motors_pos["1"]["phiz"]
        params += "%0.3f\t" % motors_pos["1"]["phiy"]
        params += "%0.3f\t" % motors_pos["1"]["sampx"]
        params += "%0.3f\t" % motors_pos["1"]["sampy"]
        params += "%0.3f\t" % motors_pos["2"]["phiz"]
        params += "%0.3f\t" % motors_pos["2"]["phiy"]
        params += "%0.3f\t" % motors_pos["2"]["sampx"]
        params += "%0.3f" % motors_pos["2"]["sampy"]

        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_scan4d",
            },
            "startScan4DEx",
        )

        scan(params)

        if wait:
            # Timeout of 15 min
            self._wait_ready(900)

    def oscilScanMesh(
        self,
        start,
        end,
        exptime,
        dead_time,
        mesh_num_lines,
        mesh_total_nb_frames,
        mesh_center,
        mesh_range,
        wait=False,
    ):
        self.scan_detector_gate_pulse_enabled.set_value(True)
        dead_time = HWR.beamline.detector.get_deadtime()
        self.scan_detector_gate_pulse_readout_time.set_value(dead_time * 1000)
        self.move_motors(mesh_center.as_dict())
        positions = self.get_positions()

        params = "%0.3f\t" % (end - start)
        params += "%0.3f\t" % mesh_range["vertical_range"]
        params += "%0.3f\t" % (-mesh_range["horizontal_range"])
        params += "%0.3f\t" % start
        params += "%0.3f\t" % positions["phiz"]
        params += "%0.3f\t" % positions["phiy"]
        params += "%0.3f\t" % positions["sampx"]
        params += "%0.3f\t" % positions["sampy"]
        params += "%d\t" % mesh_num_lines
        params += "%d\t" % (mesh_total_nb_frames / mesh_num_lines)
        params += "%0.3f\t" % (exptime / mesh_num_lines)
        params += "%r\t" % True
        params += "%r\t" % True
        params += "%r\t" % True

        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_raster_scan",
            },
            "startRasterScanEx",
        )

        scan(params)

        if wait:
            # Timeout of 30 min
            self._wait_ready(1800)

    def timeresolved_mesh_scan(self, timeout=900):
        """Start timeresolved mesh.
        Args:
            timeout(float): Timeout to wait the execution of the scan [s].
                            Default value is 15 min
        """
        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_timeresolved_mesh",
            },
            "startContinuousTimeResolvedRasterScan",
        )
        scan()

        # wait for the execution of the scan
        self._wait_ready(timeout)

    def get_timeresolved_mesh_nbframes(self):
        """Get the calculated number of frames for the continuous timeresolved
           mesh scan.
        Returns:
            (int): Number of frames.
        """
        cmd = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "timeresolved_mesh_nbframes",
            },
            "ContinuousTimeResolvedRasterScanNbFrames",
        )
        return cmd.get_value()

    def get_timeresolved_mesh_exptime(self):
        """Get the calculated exposure time for the continuous timeresolved
           mesh scan.
        Returns:
            (float): Exposure time [s].
        """

        cmd = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "timeresolved_mesh_exptime",
            },
            "ContinuousTimeResolvedRasterScanExposureTime",
        )
        return cmd.get_value()

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )

        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        dx = (x - beam_pos_x) / self.pixelsPerMmY
        dy = (y - beam_pos_y) / self.pixelsPerMmZ

        phi_angle = math.radians(
            self.centringPhi.direction * self.centringPhi.get_value()
        )

        sampx = -self.centringSamplex.direction * self.centringSamplex.get_value()
        sampy = self.centringSampley.direction * self.centringSampley.get_value()

        phiy = -self.centringPhiy.direction * self.centringPhiy.get_value()
        phiz = self.centringPhiz.direction * self.centringPhiz.get_value()

        rotMatrix = numpy.matrix(
            [
                [math.cos(phi_angle), -math.sin(phi_angle)],
                [math.sin(phi_angle), math.cos(phi_angle)],
            ]
        )
        invRotMatrix = numpy.array(rotMatrix.I)

        dsampx, dsampy = numpy.dot(numpy.array([dx, dy]), invRotMatrix)

        chi_angle = math.radians(-self.chiAngle)
        chiRot = numpy.matrix(
            [
                [math.cos(chi_angle), -math.sin(chi_angle)],
                [math.sin(chi_angle), math.cos(chi_angle)],
            ]
        )

        sx, sy = numpy.dot(numpy.array([dsampx, dsampy]), numpy.array(chiRot))

        sampx = sampx + sx
        sampy = sampy + sy
        phiz = phiz + dy

        return {
            "phi": self.centringPhi.get_value(),
            "phiz": float(phiz),
            "phiy": float(phiy),
            "sampx": float(sampx),
            "sampy": float(sampy),
        }

    def motor_positions_to_screen(self, centred_positions_dict):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )

        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0

        phi_angle = math.radians(
            self.centringPhi.direction * self.centringPhi.get_value()
        )
        sampx = self.centringSamplex.direction * (
            centred_positions_dict["sampx"] - self.centringSamplex.get_value()
        )
        sampy = self.centringSampley.direction * (
            centred_positions_dict["sampy"] - self.centringSampley.get_value()
        )
        phiy = self.centringPhiy.direction * (
            centred_positions_dict["phiy"] - self.centringPhiy.get_value()
        )
        phiz = self.centringPhiz.direction * (
            centred_positions_dict["phiz"] - self.centringPhiz.get_value()
        )

        rotMatrix = numpy.matrix(
            [
                math.cos(phi_angle),
                -math.sin(phi_angle),
                math.sin(phi_angle),
                math.cos(phi_angle),
            ]
        )
        rotMatrix.shape = (2, 2)
        invRotMatrix = numpy.array(rotMatrix.I)

        dsx, dsy = numpy.dot(numpy.array([sampx, sampy]), invRotMatrix)

        chi_angle = math.radians(self.chiAngle)
        chiRot = numpy.matrix(
            [
                math.cos(chi_angle),
                -math.sin(chi_angle),
                math.sin(chi_angle),
                math.cos(chi_angle),
            ]
        )
        chiRot.shape = (2, 2)

        sx, sy = numpy.dot(numpy.array([0, dsy]), numpy.array(chiRot))

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        x = (sx + phiy) * self.pixelsPerMmY + beam_pos_x
        y = (sy + phiz) * self.pixelsPerMmZ + beam_pos_y

        return float(x), float(y)

    def move_to_beam(self, x, y):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )

        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        dx = (x - beam_pos_x) / self.pixelsPerMmY
        dy = (y - beam_pos_y) / self.pixelsPerMmZ

        phi_angle = math.radians(
            self.centringPhi.direction * self.centringPhi.get_value()
        )

        sampx = -self.centringSamplex.direction * self.centringSamplex.get_value()
        sampy = self.centringSampley.direction * self.centringSampley.get_value()
        phiz = self.centringPhiz.direction * self.centringPhiz.get_value()

        rotMatrix = numpy.matrix(
            [
                [math.cos(phi_angle), -math.sin(phi_angle)],
                [math.sin(phi_angle), math.cos(phi_angle)],
            ]
        )
        invRotMatrix = numpy.array(rotMatrix.I)

        dsampx, dsampy = numpy.dot(numpy.array([dx, 0]), invRotMatrix)

        chi_angle = math.radians(-self.chiAngle)
        chiRot = numpy.matrix(
            [
                [math.cos(chi_angle), -math.sin(chi_angle)],
                [math.sin(chi_angle), math.cos(chi_angle)],
            ]
        )

        sx, sy = numpy.dot(numpy.array([dsampx, dsampy]), numpy.array(chiRot))

        sampx = sampx + sx
        sampy = sampy + sy
        phiz = phiz + dy

        try:
            self.centringSamplex.set_value(sampx)
            self.centringSampley.set_value(sampy)
            self.centringPhiz.set_value(phiz)
        except Exception:
            msg = "MiniDiff: could not center to beam, aborting"
            logging.getLogger("HWR").exception(msg)



try:
    from SpecClient_gevent import (
        SpecConnectionsManager,
        SpecEventsDispatcher,
    )
except ImportError:
    from SpecClient import SpecEventsDispatcher
    from SpecClient import SpecConnectionsManager

from mxcubecore.BaseHardwareObjects import Procedure


class SpecScan(Procedure):
    def __init__(self, name):
        Procedure.__init__(self, name)

        self.specConnection = None

    def setSpecVersion(self, specVersion):
        scanCmd = self.get_command_object("start")
        scanCmd.setSpecVersion(specVersion)

        if specVersion is not None:
            self.specConnection = (
                SpecConnectionsManager.SpecConnectionsManager().get_connection(
                    specVersion
                )
            )

            if self.specConnection is not None:
                SpecEventsDispatcher.connect(
                    self.specConnection, "connected", self.connected
                )
                SpecEventsDispatcher.connect(
                    self.specConnection, "disconnected", self.disconnected
                )

                if self.specConnection.isSpecConnected():
                    self.connected()

    def connected(self):
        pass

    def is_connected(self):
        return self.specConnection is not None and self.specConnection.isSpecConnected()

    def disconnected(self):
        pass

    def scanDimension(self):
        return self.get_property("dimension")

    def isAbsolute(self):
        raise NotImplementedError

    def isRelative(self):
        return not self.isAbsolute()

    def allowDifferentNbPoints(self):
        return False

    def abortScan(self):
        scanCmd = self.get_command_object("start")
        scanCmd.abort()

    def startScan(self, *args):
        scanCmd = self.get_command_object("start")
        scanCmd.connect_signal("commandReplyArrived", self.scanDone)
        scanCmd.connect_signal("commandFailed", self.scanDone)
        scanCmd.connect_signal("commandBeginWaitReply", self.scanStarted)
        scanCmd.connect_signal("commandAborted", self.scanAborted)

        scanCmd(*args)

    def scanDone(self):
        self.emit("scanDone", ())

    def scanStarted(self):
        self.emit("scanStarted", ())

    def scanAborted(self):
        self.emit("scanAborted", ())



"""Class

template:
  <procedure class="SpecShell">
    <specversion>spec host:spec name</specversion>
  </procedure>
"""

import logging

from mxcubecore.BaseHardwareObjects import HardwareObject

try:
    import SpecClient_gevent as SpecClient
except ImportError:
    import SpecClient

from qt import (
    PYSIGNAL,
    QObject,
)


class SpecOutputVar(QObject, SpecClient.SpecVariable.SpecVariableA):
    def __init__(self, parent):
        QObject.__init__(self, parent)
        SpecClient.SpecVariable.SpecVariableA.__init__(self)

    def update(self, value):
        value = str(value).rstrip()
        if len(value):
            self.emit(PYSIGNAL("outputReceived"), (value,))


class SpecShell(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)
        self.isSpecReady = False

    def init(self):
        self.specShellCommand = None
        self.specShellLsdef = None
        self.commandRunning = False
        self.lsdefRunning = False
        self.lsdefBuffer = []
        self.lsdefCommands = None
        self.specOutput = SpecOutputVar(None)
        QObject.connect(
            self.specOutput, PYSIGNAL("outputReceived"), self.outputReceived
        )
        try:
            self.specConnection = SpecClient.SpecConnectionsManager.SpecConnectionsManager().get_connection(
                self.specversion
            )
        except AttributeError:
            self.specConnection = None
            logging.getLogger("HWR").error("SpecShell: you must specify a spec version")
        else:
            self.specOutput.connectToSpec(
                "output/tty",
                self.specversion,
                dispatchMode=SpecClient.SpecVariable.FIREEVENT,
                prefix=False,
            )

            SpecClient.SpecEventsDispatcher.connect(
                self.specConnection, "connected", self.sConnected
            )
            SpecClient.SpecEventsDispatcher.connect(
                self.specConnection, "disconnected", self.sDisconnected
            )
            if self.is_connected():
                self.sConnected()

    def sConnected(self):
        self.emit("connected", ())

        speccommand = SpecClient.SpecCommand.SpecCommand(
            "sleep", self.specConnection, 0
        )

        self.add_command(
            {"name": "SpecShellMacro", "type": "spec", "version": self.specversion},
            "sleep",
        )
        cmd = self.get_command_object("SpecShellMacro")
        cmd.connect_signal("commandReady", self.commandReady)
        cmd.connect_signal("commandNotReady", self.commandNotReady)
        cmd.connect_signal("commandReplyArrived", self.commandFinished)
        cmd.connect_signal("commandBeginWaitReply", self.commandStarted)
        cmd.connect_signal("commandFailed", self.commandFailed)
        cmd.connect_signal("commandAborted", self.commandAborted)
        self.specShellCommand = cmd

        self.add_command(
            {"name": "SpecShellLsdef", "type": "spec", "version": self.specversion},
            "lsdef *",
        )
        cmd = self.get_command_object("SpecShellLsdef")
        # cmd.connect_signal('commandReady',self.commandReady)
        # cmd.connect_signal('commandNotReady',self.commandNotReady)
        cmd.connect_signal("commandReplyArrived", self.commandFinished)
        cmd.connect_signal("commandBeginWaitReply", self.commandStarted)
        cmd.connect_signal("commandFailed", self.commandFailed)
        cmd.connect_signal("commandAborted", self.commandAborted)
        self.specShellLsdef = cmd

        """
        try:
            speccommand.executeCommand("sleep(0)")
        except SpecClient.SpecClientError.SpecClientError,diag:
            pass
        """

    def is_connected(self):
        return self.specConnection is not None and self.specConnection.isSpecConnected()

    def is_ready(self):
        return self.isSpecReady

    def isRunning(self):
        return self.commandRunning

    def sDisconnected(self):
        if self.commandRunning:
            self.emit("aborted", ())
            self.emit("failed", (None,))
            self.commandRunning = False

        self.specShellCommand = None
        self.lsdefCommands = None
        self.emit("disconnected", ())

    def commandReady(self):
        self.isSpecReady = True
        if self.specShellCommand is None:
            return
        if self.commandRunning:
            return
        self.emit("ready", ())

    def commandNotReady(self):
        self.isSpecReady = False
        if self.specShellCommand is None:
            return
        if self.commandRunning:
            return
        self.emit("busy", ())

    def commandFinished(self, result, command):
        if command == "SpecShellLsdef":
            self.lsdefRunning = False
            commands_list = []
            for buf in self.lsdefBuffer:
                try:
                    buf_list = buf.split()
                except Exception:
                    pass
                else:
                    i = 0
                    while i < len(buf_list):
                        cmd_name = buf_list[i]
                        try:
                            cmd_aux = buf_list[i + 1]
                        except Exception:
                            pass
                        else:
                            try:
                                left_par = cmd_aux[0]
                                right_par = cmd_aux[-1]
                                midle_num = cmd_aux[1:-1]
                            except Exception:
                                pass
                            else:
                                if left_par == "(" and right_par == ")":
                                    try:
                                        int(midle_num)
                                    except Exception:
                                        pass
                                    else:
                                        commands_list.append(cmd_name.lstrip("*"))
                        i += 2
            self.lsdefBuffer = []
            commands_list.sort()
            self.lsdefCommands = commands_list
            self.emit("allCommandsList", (commands_list,))
        self.commandRunning = False
        self.emit("finished", (result,))

    def commandStarted(self, command):
        if command == "SpecShellLsdef":
            self.lsdefBuffer = []
            self.lsdefRunning = True
        self.commandRunning = True
        self.emit("started", ())

    def commandFailed(self, result, command):
        if command == "SpecShellLsdef":
            self.lsdefRunning = False
            self.lsdefBuffer = []
            self.lsdefCommands = None
            self.emit("allCommandsList", ((),))
        self.commandRunning = False
        self.emit("failed", (result,))

    def commandAborted(self, command):
        if command == "SpecShellLsdef":
            self.lsdefRunning = False
            self.lsdefBuffer = []
            # self.emit('allCommandsList', ((),))
        self.commandRunning = False
        self.emit("aborted", ())

    def execute_command(self, command):
        try:
            self.specShellCommand.executeCommand(command)
        except SpecClient.SpecClientError.SpecClientError as diag:
            self.emit("failed", (None,))

    def abortCommand(self):
        if self.commandRunning:
            try:
                self.specShellCommand.abort()
            except SpecClient.SpecClientError.SpecClientError as diag:
                pass

    def outputReceived(self, output):
        if self.lsdefRunning:
            self.lsdefBuffer.append(output)
        else:
            self.emit("output", (output,))

    def getSpecVersion(self):
        try:
            ver = self.specversion
        except AttributeError:
            ver = None
        return ver

    def getUserCommands(self):
        cmds = []
        try:
            for cmd in self["usercmds"]:
                try:
                    cmd_args = cmd.args
                except Exception:
                    cmd_args = ""
                if len(cmd_args):
                    cmds.append("%s %s" % (cmd.method, cmd_args))
                else:
                    cmds.append(cmd.method)
        except Exception:
            pass
        return cmds

    def getAllCommands(self):
        if self.lsdefRunning:
            return
        if self.lsdefCommands is not None:
            self.emit("allCommandsList", (self.lsdefCommands,))
        else:
            try:
                self.specShellLsdef.executeCommand("lsdef *")
            except SpecClient.SpecClientError.SpecClientError as diag:
                self.emit("allCommandsList", ((),))



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


import time

import numpy

from mxcubecore.HardwareObjects.abstract.AbstractOnlineProcessing import (
    AbstractOnlineProcessing,
)

__license__ = "LGPLv3"


class OnlineProcessingMockup(AbstractOnlineProcessing):
    def __init__(self, name):
        AbstractOnlineProcessing.__init__(self, name)

        self.result_plot_type = None

    def init(self):
        AbstractOnlineProcessing.init(self)
        self.result_plot_type = self.get_property("result_plot_type", "random")

    def run_processing(self, data_collection):
        """
        no     : no frames with a score
        firste : first frame has a score
        middle : middle frame has a score
        last   : last frame has a score
        linear : score distributed in a linear way
        random : random score distribution
        """
        self.data_collection = data_collection
        self.prepare_processing()
        self.started = True

        index = 0
        for key in self.results_raw.keys():
            if self.result_plot_type == "first":
                self.results_raw[key][0] = 1
            elif self.result_plot_type == "last":
                self.results_raw[key][self.params_dict["images_num"] - 1] = 1
            elif self.result_plot_type == "middle":
                self.results_raw[key][self.params_dict["images_num"] / 2 - 1] = 1
                self.results_raw[key][self.params_dict["images_num"] / 2] = 3
                self.results_raw[key][self.params_dict["images_num"] / 2 + 1] = 2.5
            elif self.result_plot_type == "linear":
                self.results_raw[key] = (
                    numpy.linspace(
                        0,
                        self.params_dict["images_num"],
                        self.params_dict["images_num"],
                    )
                    + index
                )
            elif self.result_plot_type == "random":
                self.results_raw[key] = numpy.random.randint(
                    1, self.params_dict["images_num"], self.params_dict["images_num"]
                )

            if key == "spots_resolution":
                self.results_raw[key] = (
                    self.results_raw[key]
                    / float(self.params_dict["images_num"])
                    * self.params_dict["resolution"]
                    + self.params_dict["resolution"]
                )
                self.results_raw[key] = 1.0 / self.results_raw[key]

            index += 1

        self.emit(
            "processingStarted",
            (self.params_dict, self.results_raw, self.results_aligned),
        )

        step = 10
        for index in range(self.params_dict["images_num"]):
            if index > 0 and not index % step:
                self.align_processing_results(index - step, index)
                self.print_log(
                    "GUI",
                    "info",
                    "Parallel processing: Frame %d/%d done"
                    % (index + 1, self.params_dict["images_num"]),
                )
                self.emit("processingFrame", index)
                self.emit("processingResultsUpdate", False)
            if not self.started:
                break
            else:
                time.sleep(self.params_dict["exp_time"])
        self.align_processing_results(0, self.params_dict["images_num"] - 1)
        self.emit("processingResultsUpdate", True)
        self.set_processing_status("Success")



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
from time import sleep

import gevent
import tine
from scipy.interpolate import interp1d

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__category__ = "General"


class EMBLBeamCentering(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.ready_event = None
        self.scale_hor = None
        self.scale_ver = None
        self.scale_double_hor = None
        self.scale_double_ver = None
        self.scan_status = None

        self.chan_pitch_scan_status = None
        # self.chan_qbpm_ar = None
        self.chan_pitch_position_ar = None
        self.cmd_set_pitch_position = None
        self.cmd_set_pitch = None
        self.cmd_start_pitch_scan = None
        self.cmd_set_vmax_pitch = None
        self.cmd_set_qbmp_range = None

        self.crl_hwobj = None
        self.beam_focusing_hwobj = None
        self.horizontal_motor_hwobj = None
        self.vertical_motor_hwobj = None
        self.horizontal_double_mode_motor_hwobj = None
        self.vertical_double_mode_motor_hwobj = None

    def init(self):
        """Reads config xml, initiates all necessary hwobj, channels and cmds
        """
        self.ready_event = gevent.event.Event()

        self.scale_hor = self.get_property("scale_hor")
        self.scale_ver = self.get_property("scale_ver")
        self.scale_double_hor = self.get_property("scale_double_hor")
        self.scale_double_ver = self.get_property("scale_double_ver")
        self.chan_pitch_scan_status = self.get_channel_object("chanPitchScanStatus")
        self.connect(
            self.chan_pitch_scan_status, "update", self.pitch_scan_status_changed
        )

        # self.chan_qbpm_ar = self.get_channel_object("chanQBPMAr")

        self.chan_pitch_position_ar = self.get_channel_object("chanPitchPositionAr")
        self.cmd_set_pitch_position = self.get_command_object("cmdSetPitchPosition")
        self.cmd_set_pitch = self.get_command_object("cmdSetPitch")
        self.cmd_start_pitch_scan = self.get_command_object("cmdStartPitchScan")
        self.cmd_set_vmax_pitch = self.get_command_object("cmdSetVMaxPitch")
        self.cmd_set_qbmp_range = self.get_command_object("cmdQBPMRangeSet")

        self.horizontal_motor_hwobj = self.get_object_by_role("horizontal_motor")
        self.vertical_motor_hwobj = self.get_object_by_role("vertical_motor")
        self.horizontal_double_mode_motor_hwobj = self.get_object_by_role(
            "horizontal_double_mode_motor"
        )
        self.vertical_double_mode_motor_hwobj = self.get_object_by_role(
            "vertical_double_mode_motor"
        )

        # self.chan_pitch_second = self.get_channel_object("chanPitchSecond")
        self.crl_hwobj = self.get_object_by_role("crl")
        self.connect(HWR.beamline.energy, "beamAlignmentRequested", self.center_beam)

        if hasattr(HWR.beamline.beam, "beam_focusing_hwobj"):
            self.beam_focusing_hwobj = HWR.beamline.beam.beam_focusing_hwobj
            self.connect(
                self.beam_focusing_hwobj,
                "focusingModeChanged",
                self.focusing_mode_changed,
            )
        else:
            logging.getLogger("HWR").debug(
                "BeamlineTest: Beam focusing hwobj is not defined"
            )

    def focusing_mode_changed(self, focusing_mode, beam_size):
        """Reemits focusing changed signal

        :param focusing_mode: focusing mode
        :type focusing_mode: str
        :param beam_size: beam size in microns
        :type beam_size: list with two int
        """
        self.emit("focusingModeChanged", focusing_mode, beam_size)

    def get_focus_mode(self):
        """Returns active focusing mode"""
        if self.beam_focusing_hwobj is not None:
            return self.beam_focusing_hwobj.get_active_focus_mode()
        else:
            return "Collimated", None

    def pitch_scan(self):
        self.cmd_start_pitch_scan(1)
        sleep(3)
        with gevent.Timeout(20, Exception("Timeout waiting for pitch scan ready")):
            while self.scan_status != 0:  # chan_pitch_scan_status.get_value() != 0:
                gevent.sleep(0.1)
                logging.getLogger("HWR").error("scan status %s" % self.scan_status)
        self.cmd_set_vmax_pitch(1)
        sleep(3)

    def center_beam(self):
        """Calls gevent task to center beam"""
        gevent.spawn(self.center_beam_task)

    def center_beam_task(self):
        """Centers beam in a following procedure:
            1. Store aperture position and take out the aperture
            2. Store slits position and open to max
            3. Do pitch scan if possible
            3. In a loop take snapshot and move motors
            4. Put back aperture and move to original slits positions
        """
        gui_log = logging.getLogger("GUI")
        log_msg = ""

        if not HWR.beamline.safety_shutter.is_opened():
            log_msg = "Beam centering failed! Safety shutter is closed! Open the shutter to continue."

            gui_log.error(log_msg)
            self.ready_event.set()
            return

        aperture_hwobj = HWR.beamline.beam.aperture
        current_energy = HWR.beamline.energy.get_value()
        current_transmission = HWR.beamline.transmission.get_value()
        active_mode, beam_size = self.get_focus_mode()

        log_msg = "Beam centering: Active mode %s" % active_mode
        gui_log.info(log_msg)

        if active_mode in ("Imaging", "TREXX"):

            log_msg = "Beam centering: doing pitch scan only"
            gui_log.info(log_msg)

            if current_energy < 10:
                crl_value = self.crl_hwobj.get_crl_value()
                self.crl_hwobj.set_crl_value([1, 1, 1, 1, 1, 1], timeout=30)

            self.cmd_start_pitch_scan(1)
            gevent.sleep(2.0)

            with gevent.Timeout(
                10, RuntimeError("Timeout waiting for pitch scan ready")
            ):
                while self.chan_pitch_scan_status.get_value() != 0:
                    gevent.sleep(0.1)
            self.cmd_set_vmax_pitch(1)

            if current_energy < 10:
                self.crl_hwobj.set_crl_value(crl_value, timeout=30)
                sleep(2)
            gui_log.info("Beam centering: done")
            self.ready_event.set()
            return

        try:
            step = 1
            log_msg = "Starting beam centring"
            gui_log.info("Beam centering: %s" % log_msg)
            self.emit("progressInit", ("Beam centering...", 20, True))

            # Diffractometer in BeamLocation phase ---------------------------
            msg = "Setting diffractometer in BeamLocation phase"
            gui_log.info("Beam centering: %s" % msg)
            self.emit("progressStep", step, msg)

            HWR.beamline.diffractometer.wait_device_ready(10)
            HWR.beamline.diffractometer.set_phase(
                HWR.beamline.diffractometer.PHASE_BEAM, timeout=45
            )

            # Open the fast shutter and set aperture out  --------------------

            step += 1
            log_msg = "Opening fast shutter and setting aperture out"
            gui_log.info("Beam centering: %s" % log_msg)
            self.emit("progressStep", step, log_msg)

            HWR.beamline.fast_shutter.openShutter()
            gevent.sleep(0.1)
            aperture_hwobj.set_out()

            # Adjust transmission ---------------------------------------------
            step += 1
            log_msg = (
                "Adjusting transmission to the current energy %.1f keV" % current_energy
            )
            gui_log.info("Beam centering: %s" % log_msg)
            self.emit("progressStep", step, log_msg)

            if current_energy < 7:
                new_transmission = 100
            else:
                energy_transm = interp1d(
                    [6.9, 8.0, 12.7, 19.0], [100.0, 60.0, 15.0, 10]
                )
                new_transmission = round(energy_transm(current_energy).tolist(), 2)

            if HWR.beamline.session.beamline_name == "P13":
                HWR.beamline.transmission.set_value(  # Transmission(
                    new_transmission, timeout=45
                )

                # TODO re
                #HWR.beamline.diffractometer.set_zoom(
                #    "Zoom 4"
                #)

                # was 4, use 1 with broken zoom motor
                # capillary_position = (
                #    HWR.beamline.diffractometer.get_capillary_position()
                # )
                HWR.beamline.diffractometer.set_capillary_position("OFF")

                gevent.sleep(1)
                self.move_beam_to_center()
            else:
                slits_hwobj = HWR.beamline.beam.slits

                if active_mode in ("Collimated", "Imaging", "TREXX"):
                    HWR.beamline.transmission.set_value(  # Transmission(
                        new_transmission, timeout=45
                    )
                    #HWR.beamline.diffractometer.set_zoom("Zoom 4")
                else:
                    # 2% transmission for beam centering in double foucused mode
                    HWR.beamline.transmission.set_value(2, timeout=45)
                    #HWR.beamline.diffractometer.set_zoom("Zoom 8")

                step += 1
                log_msg = "Opening slits to 1 x 1 mm"
                gui_log.info("Beam centering: %s" % log_msg)
                self.emit("progressStep", step, log_msg)

                # GB: keep standard slits settings for double foucsed mode
                if active_mode in ("Collimated", "Imaging", "TREXX"):
                    slits_hwobj.set_vertical_gap(1.0)  # "Hor", 1.0)
                    slits_hwobj.set_horizontal_gap(1.0)  # "Ver", 1.0)

                # Actual centring procedure  ---------------

                result = self.move_beam_to_center()
                if not result:
                    gui_log.error("Beam centering: Failed")
                    self.emit("progressStop", ())
                    self.ready_event.set()
                    return

                # For unfocused mode setting slits to 0.1 x 0.1 mm ---------------
                if active_mode in ("Collimated", "Imaging", "TREXX"):
                    step += 1
                    log_msg = "Setting slits to 0.1 x 0.1 mm"
                    gui_log.info("Beam centering: %s" % log_msg)
                    self.emit("progressStep", step, log_msg)

                    slits_hwobj.set_horizontal_gap(0.1)  # "Hor", 0.1)
                    slits_hwobj.set_vertical_gap(0.1)  # "Ver", 0.1)
                    sleep(3)

                # Update position of the beam mark position ----------------------
                step += 1
                log_msg = "Updating beam mark position"
                self.emit("progressStep", step, log_msg)
                gui_log.info("Beam centering: %s" % log_msg)

                HWR.beamline.sample_view.move_beam_mark_auto()

            HWR.beamline.transmission.set_value(current_transmission)
            HWR.beamline.sample_view.graphics_beam_item.set_detected_beam_position(
                None, None
            )

            log_msg = "Done"
            gui_log.info("Beam centering: %s" % log_msg)
            self.emit("progressStop", ())
            self.ready_event.set()

        except Exception as ex:
            log_msg = "Beam centering failed in the step: %s (%s)" % (log_msg, str(ex))
            gui_log.error(log_msg)
            self.emit("progressStop", ())
            self.ready_event.set()
            return False
        finally:
            HWR.beamline.fast_shutter.closeShutter(wait=False)

    def move_beam_to_center(self):
        """Calls pitch scan and 3 times detects beam shape and
           moves horizontal and vertical motors.
        """
        gui_log = logging.getLogger("GUI")
        gui_msg = ""
        step = 10
        finished = False

        try:
            if HWR.beamline.session.beamline_name == "P13":
                # Beam centering procedure for P13 ---------------------------------

                log_msg = "Executing pitch scan"
                gui_log.info("Beam centering: %s" % log_msg)
                self.emit("progressStep", step, log_msg)

                if HWR.beamline.energy.get_value() <= 8.75:
                    self.cmd_set_qbmp_range(0)
                else:
                    self.cmd_set_qbmp_range(1)

                gevent.sleep(0.2)
                self.cmd_start_pitch_scan(1)
                gevent.sleep(3)

                with gevent.Timeout(
                    20, Exception("Timeout waiting for pitch scan ready")
                ):
                    while self.chan_pitch_scan_status.get_value() == 1:
                        gevent.sleep(0.1)
                gevent.sleep(3)
                self.cmd_set_vmax_pitch(1)

                """
                qbpm_arr = self.chan_qbpm_ar.get_value()
                if max(qbpm_arr) < 10:
                    gui_log.error("Beam alignment failed! Pitch scan failed.")
                    self.emit("progressStop", ())
                    return
                """

                step += 1
                log_msg = "Detecting beam position and centering the beam"
                gui_log.info("Beam centering: %s" % log_msg)
                self.emit("progressStep", step, log_msg)

                for i in range(3):
                    with gevent.Timeout(10, False):
                        beam_pos_displacement = [None, None]
                        while None in beam_pos_displacement:
                            beam_pos_displacement = HWR.beamline.sample_view.get_beam_displacement(
                                reference="beam"
                            )
                            gevent.sleep(0.1)
                    if None in beam_pos_displacement:
                        log_msg = (
                            "Beam alignment failed! Unable to detect beam position."
                        )
                        gui_log.error(log_msg)
                        self.emit("progressStop", ())
                        return

                    delta_hor = beam_pos_displacement[0] * self.scale_hor
                    delta_ver = beam_pos_displacement[1] * self.scale_ver

                    if delta_hor > 0.03:
                        delta_hor = 0.03
                    if delta_hor < -0.03:
                        delta_hor = -0.03
                    if delta_ver > 0.03:
                        delta_ver = 0.03
                    if delta_ver < -0.03:
                        delta_ver = -0.03

                    log_msg = (
                        "Beam centering: Applying %.4f mm horizontal " % delta_hor
                        + "and %.4f mm vertical correction" % delta_ver
                    )
                    gui_log.info(log_msg)

                    if abs(delta_hor) > 0.0001:
                        log_msg = (
                            "Beam centering: Moving horizontal by %.4f" % delta_hor
                        )
                        gui_log.info(log_msg)
                        self.horizontal_motor_hwobj.set_value_relative(delta_hor)
                        sleep(5)
                    if abs(delta_ver) > 0.0001:
                        log_msg = "Beam centering: Moving vertical by %.4f" % delta_ver
                        gui_log.info(log_msg)
                        self.vertical_motor_hwobj.set_value_relative(delta_ver)
                        sleep(5)

            else:
                # Beam centering procedure for P14 -----------------------------------
                # If energy < 10: set all lenses in ----------------------------
                active_mode, beam_size = self.get_focus_mode()
                log_msg = "Applying Perp and Roll2nd correction"
                gui_log.info("Beam centering: %s" % log_msg)
                self.emit("progressStep", step, log_msg)
                delta_ver = 1.0

                for i in range(5):
                    if abs(delta_ver) > 0.100:
                        self.cmd_set_pitch_position(0)
                        self.cmd_set_pitch(1)
                        gevent.sleep(0.1)

                        if HWR.beamline.energy.get_value() < 10:
                            crl_value = self.crl_hwobj.get_crl_value()
                            self.crl_hwobj.set_crl_value([1, 1, 1, 1, 1, 1], timeout=30)

                        self.cmd_start_pitch_scan(1)

                        # GB : keep lenses in the beam during the scan
                        # if self.bl_hwobj._get_energy() < 10:
                        #   self.crl_hwobj.set_crl_value(crl_value, timeout=30)

                        gevent.sleep(2.0)

                        with gevent.Timeout(
                            10, RuntimeError("Timeout waiting for pitch scan ready")
                        ):
                            while self.chan_pitch_scan_status.get_value() != 0:
                                gevent.sleep(0.1)
                        self.cmd_set_vmax_pitch(1)

                        # GB : return original lenses only after scan finished
                        if HWR.beamline.energy.get_value() < 10:
                            self.crl_hwobj.set_crl_value(crl_value, timeout=30)
                        sleep(2)

                    with gevent.Timeout(10, False):
                        beam_pos_displacement = [None, None]
                        while None in beam_pos_displacement:
                            beam_pos_displacement = HWR.beamline.sample_view.get_beam_displacement(
                                reference="screen"
                            )
                            gevent.sleep(0.1)
                    if None in beam_pos_displacement:
                        # log.debug("No beam detected")
                        return
                    if active_mode in ("Collimated", "Imaging"):
                        delta_hor = (
                            beam_pos_displacement[0]
                            * self.scale_hor
                            * HWR.beamline.energy.get_value()
                            / 12.70
                        )
                        delta_ver = beam_pos_displacement[1] * self.scale_ver
                    else:
                        delta_hor = beam_pos_displacement[0] * self.scale_double_hor
                        delta_ver = (
                            beam_pos_displacement[1] * self.scale_double_ver * 0.5
                        )

                    log_msg = (
                        "Measured beam displacement: Horizontal "
                        + "%.4f mm, Vertical %.4f mm" % beam_pos_displacement
                    )
                    gui_log.info(log_msg)

                    # if abs(delta_ver) > 0.050 :
                    #    delta_ver *= 0.5

                    log_msg = (
                        "Applying %.4f mm horizontal " % delta_hor
                        + "and %.4f mm vertical motor correction" % delta_ver
                    )
                    gui_log.info(log_msg)

                    if active_mode in ("Collimated", "Imaging"):
                        if abs(delta_hor) > 0.0001:
                            log_msg = "Moving horizontal by %.4f" % delta_hor
                            gui_log.info(log_msg)
                            self.horizontal_motor_hwobj.set_value_relative(
                                delta_hor, timeout=5
                            )
                            sleep(4)
                        if abs(delta_ver) > 0.100:
                            log_msg = "Moving vertical motor by %.4f" % delta_ver
                            gui_log.info(log_msg)
                            # self.vertical_motor_hwobj.set_value_relative(delta_ver, timeout=5)
                            tine.set(
                                "/p14/P14MonoMotor/Perp",
                                "IncrementMove.START",
                                delta_ver * 0.5,
                            )
                            sleep(6)
                        else:
                            log_msg = "Moving vertical piezo by %.4f" % delta_ver
                            gui_log.info(log_msg)
                            self.vertical_motor_hwobj.set_value_relative(
                                -1.0 * delta_ver, timeout=5
                            )
                            sleep(2)

                    elif active_mode == "Double":
                        if abs(delta_hor) > 0.0001:
                            log_msg = "Moving horizontal by %.4f" % delta_hor
                            gui_log.info(log_msg)
                            self.horizontal_double_mode_motor_hwobj.set_value_relative(
                                delta_hor, timeout=5
                            )
                            sleep(2)
                        if abs(delta_ver) > 0.001:
                            log_msg = "Moving vertical by %.4f" % delta_ver
                            gui_log.info(log_msg)
                            self.vertical_double_mode_motor_hwobj.set_value_relative(
                                delta_ver, timeout=5
                            )
                            sleep(2)
            finished = True
        except Exception:
            gui_log.error("Beam centering failed")
            finished = False
        finally:
            return finished

    def pitch_scan_status_changed(self, status):
        """Store pitch scan status"""
        self.scan_status = status



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Machine Current Tango Hardware Object
Example XML:
<object class = "MachCurrent">
  <username>label for users</username>
  <tangoname>orion:10000/fe/id(or d)/xx</tangoname>
  <channel type="tango" name="OperatorMsg" polling="2000">SR_Operator_Mesg</channel>
  <channel type="tango" name="Current" polling="2000">SR_Current</channel>
  <channel type="tango" name="FillingMode" polling="2000">SR_Filling_Mode</channel>
  <channel type="tango" name="RefillCountdown" polling="2000">SR_Refill_Countdown</channel>
</object>
"""

import logging

from mxcubecore.HardwareObjects.abstract.AbstractMachineInfo import AbstractMachineInfo

__copyright__ = """ Copyright © 2010-2023 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MachCurrent(AbstractMachineInfo):
    """Tango implementation"""

    def __init__(self, name):
        super().__init__(name)
        self.opmsg = ""

    def init(self):
        try:
            curr = self.get_channel_object("Current")
            curr.connect_signal("update", self.value_changed)
            self.update_state(self.STATES.READY)
        except Exception as err:
            logging.getLogger("HWR").exception(err)

    def value_changed(self, value):
        """Get information from the control software, emit valueChanged"""
        value = value or self.get_current()

        try:
            opmsg = self.get_channel_object("OperatorMsg").get_value()
            fillmode = self.get_channel_object("FillingMode").get_value()
            fillmode = fillmode.strip()

            refill = self.get_channel_object("RefillCountdown").get_value()
        except Exception as err:
            logging.getLogger("HWR").exception(err)
            opmsg, fillmode, value, refill = ("", "", -1, -1)

        if opmsg and opmsg != self.opmsg:
            self.opmsg = opmsg
            logging.getLogger("user_level_log").info(self.opmsg)
        self.emit("valueChanged", (value, opmsg, fillmode, refill))

    def get_current(self) -> float:
        """Read the ring current.
        Returns:
            (float): Ring current [mA]
        """
        try:
            return self.get_channel_object("Current").get_value()
        except Exception as err:
            logging.getLogger("HWR").exception(err)
            return -1

    def get_message(self) -> str:
        try:
            return self.get_channel_object("OperatorMsg").get_value()
        except Exception as err:
            logging.getLogger("HWR").exception(err)
            return ""

    def get_fill_mode(self) -> str:
        try:
            return self.get_channel_object("FillingMode").get_value()
        except Exception as err:
            logging.getLogger("HWR").exception(err)
            return ""



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""TangoShutter class - interface for shutter controlled by TANGO
Implements _set_value, get_value methods

Tango states are:
"UNKNOWN", "OPEN", "CLOSED", "FAULT", "MOVING", "DISABLE", "STANDBY", "RUNNING"

Example xml file:

.. code-block:: xml

 <object class = "TangoShutter">
   <username>Safety Shutter</username>
   <tangoname>ab/cd/ef</tangoname>
   <command type="tango" name="Open">Open</command>
   <command type="tango" name="Close">Close</command>
   <channel type="tango" name="State" polling="1000">State</channel>
   <values>{"OPEN": "MYOPEN", "NEWSTATE": ["MYSTATE", "BUSY"]}</values>
 </object>

In the example the <values> property contains a dictionary that redefines or
adds specific tango shutter states.
When redefining a known state, only the VALUES Enum will be updated.
When defining a new state (new key), the dictionary value should be a
list. The new state is added to both the ``VALUES`` and the ``SPECIFIC_STATES`` enumeration.

Attention:

- do not use tuples or the Python JSON parser will fail!
- make sure only double quotes (``"``) are used inside the values dictionary,
  no single quotes (``'``) are allowed!
- the second element of the list should be a standard ``HardwareObjectState`` name
  (``UNKNOWN``, ``WARNING``, ``BUSY``, ``READY``, ``FAULT``, or ``OFF``;
  see in ``BaseHardwareObjects``)!

The ``<values>`` property is optional.
"""

import json
import logging
from enum import (
    Enum,
    unique,
)

from tango import DevState

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"

hwr_logger = logging.getLogger("HWR")


@unique
class TangoShutterStates(Enum):
    """Shutter states definitions."""

    CLOSED = HardwareObjectState.READY, "CLOSED"
    OPEN = HardwareObjectState.READY, "OPEN"
    MOVING = HardwareObjectState.BUSY, "MOVING"
    DISABLE = HardwareObjectState.WARNING, "DISABLE"
    AUTOMATIC = HardwareObjectState.READY, "RUNNING"
    UNKNOWN = HardwareObjectState.UNKNOWN, "RUNNING"
    FAULT = HardwareObjectState.WARNING, "FAULT"
    STANDBY = HardwareObjectState.WARNING, "STANDBY"


class TangoShutter(AbstractShutter):
    """TANGO implementation of AbstractShutter"""

    SPECIFIC_STATES = TangoShutterStates

    def __init__(self, name: str):
        super().__init__(name)
        self.open_cmd = None
        self.close_cmd = None
        self.state_channel = None

    def init(self) -> None:
        """Initialise the predefined values"""
        super().init()
        self.open_cmd = self.get_command_object("Open")
        self.close_cmd = self.get_command_object("Close")
        self.state_channel = self.get_channel_object("State")
        self._initialise_values()
        self.state_channel.connect_signal("update", self._update_value)
        self.update_state()

    def _update_value(self, value: DevState | str) -> None:
        """Update the value.

        Args:
            value: The value reported by the state channel.
        """
        super().update_value(self.value_to_enum(str(value)))

    def _initialise_values(self) -> None:
        """Add specific tango states to VALUES and, if configured
        in the xml file, to SPECIFIC_STATES"""
        values_dict = {item.name: item.value for item in self.VALUES}
        states_dict = {item.name: item.value for item in self.SPECIFIC_STATES}
        values_dict.update(
            {
                "MOVING": "MOVING",
                "DISABLE": "DISABLE",
                "STANDBY": "STANDBY",
                "FAULT": "FAULT",
            }
        )
        try:
            config_values = json.loads(self.get_property("values"))
            for key, val in config_values.items():
                if isinstance(val, (tuple, list)):
                    values_dict.update({key: val[1]})
                    states_dict.update({key: (HardwareObjectState[val[1]], val[0])})
                else:
                    values_dict.update({key: val})
        except (ValueError, TypeError):
            hwr_logger.exception("Exception in _initialise_values()")

        self.VALUES = Enum("ValueEnum", values_dict)
        self.SPECIFIC_STATES = Enum("TangoShutterStates", states_dict)

    def get_state(self) -> HardwareObjectState:
        """Get the device state.

        Returns:
            Device state.
        """
        try:
            _state = self.get_value().name
            return self.SPECIFIC_STATES[_state].value[0]
        except (AttributeError, KeyError):
            hwr_logger.exception("Exception in get_state()")
            return self.STATES.UNKNOWN

    def get_value(self) -> TangoShutterStates:
        """Get the device value

        Returns:
            Enum member, corresponding to the 'VALUE' or UNKNOWN.
        """
        _val = str(self.state_channel.get_value())
        return self.value_to_enum(_val)

    def _set_value(self, value: TangoShutterStates) -> None:
        if value.name == "OPEN":
            self.open_cmd()
        elif value.name == "CLOSED":
            self.close_cmd()
